import functools
import inspect
import logging
import os.path as op
from time import time

import nibabel as nib
from dipy.io.streamline import save_tractogram
from dipy.io.stateful_tractogram import StatefulTractogram

try:
    from trx.trx_file_memmap import TrxFile
    from trx.io import save as save_trx
    has_trx = True
except ModuleNotFoundError:
    has_trx = False

import numpy as np

from AFQ.tasks.utils import get_fname
from AFQ.utils.path import drop_extension, write_json


# These should only be used with pimms.calc
__all__ = ["as_file", "as_fit_deriv", "as_img"]


logger = logging.getLogger('AFQ')
logger.setLevel(logging.INFO)


# get args and kwargs from function
def get_args_and_kwargs(func):
    param_dict = inspect.signature(func).parameters
    param_list = func.__code__.co_varnames[
        :func.__code__.co_argcount]
    is_param_kwarg = {
        name: name in param_dict and param_dict[name].default is
        not param_dict[name].empty for name in param_list}
    return param_list, is_param_kwarg, param_dict


# replaces *args and **kwargs with specific parameters from og_func
# so that pimms can see original parameter names after wrapping
# also adds on any args the decorator requires
# these will be extracted with extract_added_args
def has_args(og_func, needed_args):
    def _has_args(func):
        header = "def wrapper_has_args_func("
        content = "):\n    return func("
        found_args = []
        param_list, is_param_kwarg, param_dict = get_args_and_kwargs(og_func)

        # add func args
        for name in param_list:
            if not is_param_kwarg[name]:
                header += f"{name}, "
                content += f"{name}, "
                found_args.append(name)

        # add decorator args
        for arg in needed_args:
            if arg not in found_args:
                header += f"{arg}, "
                content += f"{arg}, "

        # add func kwargs
        for name in param_list:
            if is_param_kwarg[name]:
                default = param_dict[name].default
                if isinstance(default, str):
                    header += f"{name}='{default}', "
                else:
                    header += f"{name}={default}, "
                content += f"{name}={name}, "

        header = header[:-2]
        content = content[:-2]
        content = f"{content})"

        wrapper_has_args = header + content
        scope = {"func": func}
        exec(wrapper_has_args, scope)
        return scope['wrapper_has_args_func']
    return _has_args


# from function where needed args (like base_fname) are added,
# return length of args before added args, and the added args
def extract_added_args(func, names, args, includes=None):
    vals = []
    param_list, is_param_kwarg, _ = get_args_and_kwargs(func)
    arg_list = [param for param in param_list if not is_param_kwarg[param]]
    extra_count = 0
    for jj, name in enumerate(names):
        if includes is not None and not includes[jj]:
            vals.append(None)
            continue

        found = False
        for ii, arg_name in enumerate(arg_list):
            if arg_name == name:
                vals.append(args[ii])
                found = True
                break
        if not found:
            vals.append(args[len(arg_list) + extra_count])
            extra_count = extra_count + 1

    return len(arg_list), *vals


def as_file(suffix, include_track=False, include_seg=False):
    """
    return img and meta as saved file path, with json,
    and only run if not already found
    """
    def _as_file(func):
        needed_args = ["base_fname", "output_dir"]
        if include_track:
            needed_args.append("tracking_params")
        if include_seg:
            needed_args.append("segmentation_params")

        @functools.wraps(func)
        @has_args(func, needed_args)
        def wrapper_as_file(*args, **kwargs):
            og_arg_count, base_fname, output_dir, \
                tracking_params, segmentation_params =\
                extract_added_args(
                    func,
                    ["base_fname", "output_dir",
                     "tracking_params", "segmentation_params"],
                    args,
                    includes=[True, True, include_track, include_seg])
            this_file = get_fname(
                base_fname, suffix,
                tracking_params=tracking_params,
                segmentation_params=segmentation_params)

            # tracking_params is defined and file has no extension
            if tracking_params is not None and not op.splitext(this_file)[1]:
                if tracking_params["trx"]:
                    this_file = this_file + ".trx"
                else:
                    this_file = this_file + ".trk"

            if not op.exists(this_file):
                gen, meta = func(*args[:og_arg_count], **kwargs)

                logger.info(f"Saving {this_file}")
                if isinstance(gen, nib.Nifti1Image):
                    nib.save(gen, this_file)
                elif isinstance(gen, StatefulTractogram):
                    save_tractogram(
                        gen, this_file, bbox_valid_check=False)
                elif isinstance(gen, np.ndarray):
                    np.save(this_file, gen)
                elif has_trx and isinstance(gen, TrxFile):
                    save_trx(gen, this_file)
                else:
                    gen.to_csv(this_file)

                # these are used to determine dependencies
                # when clobbering derivatives
                if "_desc-profiles" in suffix:
                    meta["dependent"] = "prof"
                elif include_seg:
                    meta["dependent"] = "rec"
                elif include_track:
                    meta["dependent"] = "trk"
                else:
                    meta["dependent"] = "dwi"

                # modify meta source to be relative
                if "source" in meta:
                    meta["source"] = op.relpath(meta["source"], output_dir)

                meta_fname = get_fname(
                    base_fname, f"{drop_extension(suffix)}.json",
                    tracking_params=tracking_params,
                    segmentation_params=segmentation_params)
                write_json(meta_fname, meta)
            return this_file
        return wrapper_as_file
    return _as_file


def as_fit_deriv(tf_name):
    """
    return data as nibabel image, meta with params information
    """
    def _as_fit_deriv(func):
        needed_args = ["dwi_affine", f"{tf_name.lower()}_params"]

        @functools.wraps(func)
        @has_args(func, needed_args)
        def wrapper_as_fit_deriv(*args, **kwargs):
            og_arg_count, dwi_affine, params = extract_added_args(
                func, needed_args, args)
            img = nib.Nifti1Image(
                func(*args[:og_arg_count], **kwargs), dwi_affine)
            return img, {f"{tf_name}ParamsFile": params}
        return wrapper_as_fit_deriv
    return _as_fit_deriv


def as_img(func):
    """
    return data, meta as nibabel image, meta with timing
    """
    needed_args = ["dwi_affine"]

    @functools.wraps(func)
    @has_args(func, needed_args)
    def wrapper_as_img(*args, **kwargs):
        og_arg_count, affine = extract_added_args(
            func, needed_args, args)
        start_time = time()
        data, meta = func(*args[:og_arg_count], **kwargs)
        meta['timing'] = time() - start_time
        img = nib.Nifti1Image(data.astype(np.float32), affine)
        return img, meta
    return wrapper_as_img



from collections import OrderedDict
import os.path as op
import logging

import numpy as np
import imageio as io
from PIL import Image, ImageChops

import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import matplotlib.transforms as mtransforms

import nibabel as nib
import dipy.tracking.streamlinespeed as dps
from dipy.align import resample

import AFQ.utils.volume as auv
import AFQ.registration as reg
import AFQ.utils.streamlines as aus

__all__ = ["Viz"]

viz_logger = logging.getLogger("AFQ")
tableau_20 = [
    (0.12156862745098039, 0.4666666666666667, 0.7058823529411765),
    (0.6823529411764706, 0.7803921568627451, 0.9098039215686274),
    (1.0, 0.4980392156862745, 0.054901960784313725),
    (1.0, 0.7333333333333333, 0.47058823529411764),
    (0.17254901960784313, 0.6274509803921569, 0.17254901960784313),
    (0.596078431372549, 0.8745098039215686, 0.5411764705882353),
    (0.8392156862745098, 0.15294117647058825, 0.1568627450980392),
    (1.0, 0.596078431372549, 0.5882352941176471),
    (0.5803921568627451, 0.403921568627451, 0.7411764705882353),
    (0.7725490196078432, 0.6901960784313725, 0.8352941176470589),
    (0.5490196078431373, 0.33725490196078434, 0.29411764705882354),
    (0.7686274509803922, 0.611764705882353, 0.5803921568627451),
    (0.8901960784313725, 0.4666666666666667, 0.7607843137254902),
    (0.9686274509803922, 0.7137254901960784, 0.8235294117647058),
    (0.4980392156862745, 0.4980392156862745, 0.4980392156862745),
    (0.7803921568627451, 0.7803921568627451, 0.7803921568627451),
    (0.7372549019607844, 0.7411764705882353, 0.13333333333333333),
    (0.8588235294117647, 0.8588235294117647, 0.5529411764705883),
    (0.09019607843137255, 0.7450980392156863, 0.8117647058823529),
    (0.6196078431372549, 0.8549019607843137, 0.8980392156862745)]
tableau_extension = [
    (0.95, 0.85, 0.25),  # Warm, Soft Yellow
    (0.98, 0.92, 0.5),
    (0.4, 0.5, 0.6),  # Cool, Muted Blue
    (0.6, 0.7, 0.9)]
large_font = 28
medium_font = 24
small_font = 20
marker_size = 200

COLOR_DICT = OrderedDict({
    "Left Anterior Thalamic": tableau_20[0], "C_L": tableau_20[0],
    "Right Anterior Thalamic": tableau_20[1], "C_R": tableau_20[1],
    "Left Corticospinal": tableau_20[2],
    "Right Corticospinal": tableau_20[3],
    "Left Cingulum Cingulate": tableau_20[4], "MCP": tableau_20[4],
    "Right Cingulum Cingulate": tableau_20[5], "CCMid": tableau_20[5],
    "Forceps Minor": tableau_20[8], "CC_ForcepsMinor": tableau_20[8],
    "Forceps Major": tableau_20[9], "CC_ForcepsMajor": tableau_20[9],
    "Left Inferior Fronto-occipital": tableau_20[10],
    "IFOF_L": tableau_20[10],
    "Right Inferior Fronto-occipital": tableau_20[11],
    "IFOF_R": tableau_20[11],
    "Left Inferior Longitudinal": tableau_20[12], "F_L": tableau_20[12],
    "Right Inferior Longitudinal": tableau_20[13], "F_R": tableau_20[13],
    "Left Superior Longitudinal": tableau_20[14],
    "Right Superior Longitudinal": tableau_20[15],
    "Left Uncinate": tableau_20[16], "UF_L": tableau_20[16],
    "Right Uncinate": tableau_20[17], "UF_R": tableau_20[17],
    "Left Arcuate": tableau_20[18], "AF_L": tableau_20[18],
    "Right Arcuate": tableau_20[19], "AF_R": tableau_20[19],
    "Left Posterior Arcuate": tableau_20[6],
    "Right Posterior Arcuate": tableau_20[7],
    "Left Vertical Occipital": tableau_extension[0],
    "Right Vertical Occipital": tableau_extension[1],
    "median": tableau_20[6],
    # Paul Tol's palette for callosal bundles
    "Callosum Orbital": (0.2, 0.13, 0.53),
    "Callosum Anterior Frontal": (0.07, 0.47, 0.2),
    "Callosum Superior Frontal": (0.27, 0.67, 0.6),
    "Callosum Motor": (0.53, 0.8, 0.93),
    "Callosum Superior Parietal": (0.87, 0.8, 0.47),
    "Callosum Posterior Parietal": (0.8, 0.4, 0.47),
    "Callosum Occipital": (0.67, 0.27, 0.6),
    "Callosum Temporal": (0.53, 0.13, 0.33)})

POSITIONS = OrderedDict({
    "Left Anterior Thalamic": (1, 0), "Right Anterior Thalamic": (1, 4),
    "C_L": (1, 0), "C_R": (1, 4),
    "Left Corticospinal": (1, 1), "Right Corticospinal": (1, 3),
    "Left Cingulum Cingulate": (3, 1),
    "Right Cingulum Cingulate": (3, 3),
    "MCP": (3, 1), "CCMid": (3, 3),
    "Forceps Minor": (4, 2), "Forceps Major": (0, 2),
    "CC_ForcepsMinor": (4, 2), "CC_ForcepsMajor": (0, 2),
    "Left Inferior Fronto-occipital": (4, 1),
    "Right Inferior Fronto-occipital": (4, 3),
    "IFOF_L": (4, 1), "IFOF_R": (4, 3),
    "Left Inferior Longitudinal": (3, 0),
    "Right Inferior Longitudinal": (3, 4),
    "F_L": (3, 0), "F_R": (3, 4),
    "Left Superior Longitudinal": (2, 1),
    "Right Superior Longitudinal": (2, 3),
    "Left Arcuate": (2, 0), "Right Arcuate": (2, 4),
    "AF_L": (2, 0), "AF_R": (2, 4),
    "Left Uncinate": (0, 1), "Right Uncinate": (0, 3),
    "UF_L": (0, 1), "UF_R": (0, 3)})

CSV_MAT_2_PYTHON = \
    {'fa': 'dti_fa', 'md': 'dti_md'}

SCALE_MAT_2_PYTHON = \
    {'dti_md': 0.001}

SCALAR_REMOVE_MODEL = \
    {'dti_md': 'MD', 'dki_md': 'MD', 'dki_fa': 'FA', 'dti_fa': 'FA'}

RECO_FLIP = ["IFO_L", "IFO_R", "UNC_L", "ILF_L", "ILF_R"]

BEST_BUNDLE_ORIENTATIONS = {
    "Left Anterior Thalamic": ("Sagittal", "Left"),
    "Right Anterior Thalamic": ("Sagittal", "Right"),
    "Left Corticospinal": ("Sagittal", "Left"),
    "Right Corticospinal": ("Sagittal", "Right"),
    "Left Cingulum Cingulate": ("Sagittal", "Left"),
    "Right Cingulum Cingulate": ("Sagittal", "Right"),
    "Forceps Minor": ("Axial", "Top"),
    "Forceps Major": ("Axial", "Top"),
    "Left Inferior Fronto-occipital": ("Sagittal", "Left"),
    "Right Inferior Fronto-occipital": ("Sagittal", "Right"),
    "Left Inferior Longitudinal": ("Sagittal", "Left"),
    "Right Inferior Longitudinal": ("Sagittal", "Right"),
    "Left Superior Longitudinal": ("Axial", "Top"),
    "Right Superior Longitudinal": ("Axial", "Top"),
    "Left Uncinate": ("Axial", "Bottom"),
    "Right Uncinate": ("Axial", "Bottom"),
    "Left Arcuate": ("Sagittal", "Left"),
    "Right Arcuate": ("Sagittal", "Right"),
    "Left Vertical Occipital": ("Coronal", "Back"),
    "Right Vertical Occipital": ("Coronal", "Back"),
    "Left Posterior Arcuate": ("Coronal", "Back"),
    "Right Posterior Arcuate": ("Coronal", "Back")}


class PanelFigure():
    """
    Super useful class for organizing existing images
    into subplots using matplotlib
    """

    def __init__(self, num_rows, num_cols, width, height,
                 panel_label_kwargs={}):
        """
        Initialize PanelFigure.

        Parameters
        ----------
        num_rows : int
            Number of rows in figure
        num_cols : int
            Number of columns in figure
        width : int
            Width of figure in inches
        height : int
            Height of figure in inches
        panel_label_kwargs : dict
            Additional arguments for matplotlib's text method,
            which is used to add panel labels to each subplot
        """
        self.fig = plt.figure(figsize=(width, height))
        self.grid = plt.GridSpec(num_rows, num_cols, hspace=0, wspace=0)
        self.subplot_count = 0
        self.panel_label_kwargs = dict(
            fontfamily="Helvetica-Bold",
            fontsize="xx-large",
            color="white",
            fontweight='bold',
            verticalalignment="top",
            bbox=dict(
                facecolor='none',
                edgecolor='none'))
        self.panel_label_kwargs.update(panel_label_kwargs)

    def add_img(self, fname, x_coord, y_coord, reduct_count=1,
                subplot_label_pos=(0.1, 1.0), legend=None, legend_kwargs={},
                add_panel_label=True, panel_label_kwargs={}):
        """
        Add image from fname into figure as a panel.

        Parameters
        ----------
        fname : str
            path to image file to add to subplot
        x_coord : int or slice
            x coordinate(s) of subplot in matlotlib figure
        y_coord : int or slice
            y coordinate(s) of subplot in matlotlib figure
        reduct_count : int
            number of times to trim whitespace around image
            Default: 1
        subplot_label_pos : tuple of floats
            position of subplot label
            Default: (0.1, 1.0)
        legend : dict
            dictionary of legend items, where keys are labels
            and values are colors
            Default: None
        legend_kwargs : dict
            ADditional arguments for matplotlib's legend method
        add_panel_label : bool
            Whether or not to add a panel label to the subplot
            Default: True
        panel_label_kwargs : dict
            Additional arguments for matplotlib's text method,
            which is used to add panel labels to each subplot.
        """
        ax = self.fig.add_subplot(self.grid[y_coord, x_coord])
        im1 = Image.open(fname)
        for _ in range(reduct_count):
            im1 = trim(im1)
        if legend is not None:
            patches = []
            for value, color in legend.items():
                patches.append(mpatches.Patch(
                    color=color,
                    label=value))
            ax.legend(handles=patches, borderaxespad=0., **legend_kwargs)
        if add_panel_label:
            trans = mtransforms.ScaledTranslation(
                10 / 72, -5 / 72, self.fig.dpi_scale_trans)
            this_pl_kwargs = self.panel_label_kwargs.copy()
            this_pl_kwargs.update(panel_label_kwargs)
            ax.text(
                subplot_label_pos[0], subplot_label_pos[1],
                f"{chr(65+self.subplot_count)}",
                transform=ax.transAxes + trans,
                **this_pl_kwargs)
        ax.imshow(np.asarray(im1), aspect=1)
        ax.axis('off')
        self.subplot_count = self.subplot_count + 1
        return ax

    def format_and_save_figure(self, fname, trim_final=True):
        """
        Format and save figure to fname.
        Parameters
        ----------
        fname : str
            Path to save figure to
        trim : bool
            Whether or not to trim whitespace around figure.
            Default: True
        """
        self.fig.tight_layout()
        self.fig.savefig(fname, dpi=300)
        if trim_final:
            im1 = Image.open(fname)
            im1 = trim(im1)
            im1.save(fname, dpi=(300, 300))


def get_eye(view, direc):
    direc = direc.lower()
    view = view.lower()

    if view in ["sagital", "saggital"]:
        viz_logger.warning("You don't know how to spell sagggitttal!")
        view = "sagittal"

    if view not in ["sagittal", "coronal", "axial"]:
        raise ValueError(
            "View must be one of: sagittal, coronal, or axial")

    if direc not in ["left", "right", "top", "bottom", "front", "back"]:
        raise ValueError(
            "View must be one of: left, right, top, bottom, front, back")

    eye = {}
    if view == "sagittal":
        if direc == "left":
            eye["x"] = -1
        else:
            eye["x"] = 1
        eye["y"] = 0
        eye["z"] = 0
    elif view == "coronal":
        eye["x"] = 0
        if direc == "front":
            eye["y"] = 1
        else:
            eye["y"] = -1
        eye["z"] = 0
    elif view == "axial":
        eye["x"] = 0
        eye["y"] = 0
        if direc == "top":
            eye["z"] = 1
        else:
            eye["z"] = -1
    return eye


def display_string(scalar_name):
    if isinstance(scalar_name, str):
        return scalar_name.replace("_", " ").upper()
    else:
        return [sn.replace("_", " ").upper() for sn in scalar_name]


def gen_color_dict(bundles):
    """
    Helper function.
    Generate a color dict given a list of bundles.
    """
    def incr_color_idx(color_idx):
        return (color_idx + 1) % 20
    custom_color_dict = {}
    color_idx = 0
    for bundle in bundles:
        if bundle not in custom_color_dict.keys():
            if bundle in COLOR_DICT.keys():
                custom_color_dict[bundle] = COLOR_DICT[bundle]
            else:
                other_bundle = bundle
                if bundle.startswith("Left "):
                    other_bundle = "Right" + other_bundle[5:]
                elif bundle.startswith("Right "):
                    other_bundle = "Left" + other_bundle[4:]
                other_bundle = str(other_bundle)

                if other_bundle == bundle:  # lone bundle
                    custom_color_dict[bundle] = tableau_20[color_idx]
                    color_idx = incr_color_idx(color_idx)
                else:  # right left pair
                    if color_idx % 2 != 0:
                        color_idx = incr_color_idx(color_idx)
                    custom_color_dict[bundle] =\
                        tableau_20[color_idx]
                    custom_color_dict[other_bundle] =\
                        tableau_20[color_idx + 1]
                    color_idx = incr_color_idx(incr_color_idx(color_idx))
    return custom_color_dict


def viz_import_msg_error(module):
    """Alerts user to install the appropriate viz module """
    if module == "plot":
        msg = "To make plots in pyAFQ, you will need to install "
        msg += "the relevant plotting software packages."
        msg += "You can do that by installing pyAFQ with "
        msg += "`pip install AFQ[plot]`"
    else:
        msg = f"To use {module.upper()} visualizations in pyAFQ, you will "
        msg += f"need to have {module.upper()} installed. "
        msg += "You can do that by installing pyAFQ with "
        msg += f"`pip install AFQ[{module.lower()}]`, or by "
        msg += f"separately installing {module.upper()}: "
        msg += f"`pip install {module.lower()}`."
    return msg


def tract_generator(trk_file, bundle, colors, n_points,
                    n_sls_viz=3600, n_sls_min=75):
    """
    Generates bundles of streamlines from the tractogram.
    Only generates from relevant bundle if bundle is set.
    Otherwise, returns all streamlines.

    Helper function

    Parameters
    ----------
    trk_file : str or SegmentedSFT
        Path to a trk file or SegmentedSFT

    bundle : str
        The name of a bundle to select from the trk metadata.

    colors : dict or list
        If this is a dict, keys are bundle names and values are RGB tuples.
        If this is a list, each item is an RGB tuple. Defaults to a list
        with Tableau 20 RGB values

    n_points : int or None
        n_points to resample streamlines to before plotting. If None, no
        resampling is done.

    n_sls_viz : int
        Number of streamlines to randomly select if plotting
        all bundles. Selections will be proportional to the original number of
        streamlines per bundle.
        Default: 3600
    n_sls_min : int
        Minimun number of streamlines to display per bundle.
        Default: 75

    Returns
    -------
    Statefule Tractogram streamlines, RGB numpy array, str
    """
    viz_logger.info("Loading Stateful Tractogram...")
    if isinstance(trk_file, str):
        seg_sft = aus.SegmentedSFT.fromfile(trk_file)
    else:
        seg_sft = trk_file

    if colors is None:
        colors = gen_color_dict(seg_sft.bundle_names)

    seg_sft.sft.to_vox()
    streamlines = seg_sft.sft.streamlines
    viz_logger.info("Generating colorful lines from tractography...")

    if len(seg_sft.bundle_names) == 1\
            and seg_sft.bundle_names[0] == "whole_brain":
        if isinstance(colors, dict):
            colors = list(colors.values())
        # There are no bundles in here:
        if len(streamlines) > n_sls_viz:
            idx = np.arange(len(streamlines))
            idx = np.random.choice(
                idx, size=n_sls_viz, replace=False)
            streamlines = streamlines[idx]
        if n_points is not None:
            streamlines = dps.set_number_of_points(streamlines, n_points)
        yield streamlines, colors[0], "all_bundles", seg_sft.sft.dimensions
    else:
        if bundle is None:
            # No selection: visualize all of them:
            for bundle_name in seg_sft.bundle_names:
                idx = seg_sft.bundle_idxs[bundle_name]
                if len(idx) == 0:
                    continue
                n_sl_viz = (len(idx) * n_sls_viz) //\
                    len(streamlines)
                n_sl_viz = max(n_sls_min, n_sl_viz)
                if len(idx) > n_sl_viz:
                    idx = np.random.choice(idx, size=n_sl_viz, replace=False)
                these_sls = streamlines[idx]
                if n_points is not None:
                    these_sls = dps.set_number_of_points(these_sls, n_points)
                if isinstance(colors, dict):
                    color = colors[bundle_name]
                else:
                    color = colors[0]
                yield these_sls, color, bundle_name, seg_sft.sft.dimensions
        else:
            these_sls = seg_sft.get_bundle(bundle).streamlines
            if n_points is not None:
                these_sls = dps.set_number_of_points(these_sls, n_points)
            if isinstance(colors, dict):
                color = colors[bundle]
            else:
                color = colors[0]
            yield these_sls, color, bundle, seg_sft.sft.dimensions


def bbox(img):
    img = np.sum(img, axis=-1)
    rows = np.any(img, axis=1)
    cols = np.any(img, axis=0)
    rmin, rmax = np.where(rows)[0][[0, -1]]
    cmin, cmax = np.where(cols)[0][[0, -1]]

    return cmin, rmin, cmax, rmax


def trim(im):
    bg = Image.new(im.mode, im.size, im.getpixel((0, 0)))
    diff = ImageChops.difference(im, bg)
    diff = ImageChops.add(diff, diff, 2.0, -100)
    diff.getbbox
    this_bbox = bbox(diff)
    if this_bbox:
        return im.crop(this_bbox)


def gif_from_pngs(tdir, gif_fname, n_frames,
                  png_fname="tgif", add_zeros=False):
    """
        Helper function
        Stitches together gif from screenshots
    """
    if add_zeros:
        fname_suffix10 = "00000"
        fname_suffix100 = "0000"
        fname_suffix1000 = "000"
    else:
        fname_suffix10 = ""
        fname_suffix100 = ""
        fname_suffix1000 = ""
    angles = []
    n_frame_copies = 60 // n_frames
    for i in range(n_frames):
        if i < 10:
            angle_fname = f"{png_fname}{fname_suffix10}{i}.png"
        elif i < 100:
            angle_fname = f"{png_fname}{fname_suffix100}{i}.png"
        else:
            angle_fname = f"{png_fname}{fname_suffix1000}{i}.png"
        frame = io.imread(op.join(tdir, angle_fname))
        for j in range(n_frame_copies):
            angles.append(frame)

    io.mimsave(gif_fname, angles)


def prepare_roi(roi, affine_or_mapping, static_img,
                roi_affine, static_affine, reg_template):
    """
    Load the ROI
    Possibly perform a transformation on an ROI
    Helper function

    Parameters
    ----------
    roi : str or Nifti1Image
        The ROI information.
        If str, ROI will be loaded using the str as a path.

    affine_or_mapping : ndarray, Nifti1Image, or str
       An affine transformation or mapping to apply to the ROI before
       visualization. Default: no transform.

    static_img: str or Nifti1Image
        Template to resample roi to.

    roi_affine: ndarray

    static_affine: ndarray

    reg_template: str or Nifti1Image
        Template to use for registration.

    Returns
    -------
    ndarray
    """
    viz_logger.info("Preparing ROI...")
    if not isinstance(roi, np.ndarray):
        if isinstance(roi, str):
            roi = nib.load(roi).get_fdata()
        else:
            roi = roi.get_fdata()

    if affine_or_mapping is not None:
        if isinstance(affine_or_mapping, np.ndarray):
            # This is an affine:
            if (static_img is None or roi_affine is None
                    or static_affine is None):
                raise ValueError("If using an affine to transform an ROI, "
                                 "need to also specify all of the following",
                                 "inputs: `static_img`, `roi_affine`, ",
                                 "`static_affine`")
            roi = resample(roi, static_img, roi_affine,
                           static_affine).get_fdata()
        else:
            # Assume it is  a mapping:
            if (isinstance(affine_or_mapping, str)
                    or isinstance(affine_or_mapping, nib.Nifti1Image)):
                if reg_template is None or static_img is None:
                    raise ValueError(
                        "If using a mapping to transform an ROI, need to ",
                        "also specify all of the following inputs: ",
                        "`reg_template`, `static_img`")
                affine_or_mapping = reg.read_mapping(affine_or_mapping,
                                                     static_img,
                                                     reg_template)

            roi = auv.transform_inverse_roi(
                roi,
                affine_or_mapping).astype(bool)
    return roi


def load_volume(volume):
    """
    Load a volume
    Helper function

    Parameters
    ----------
    volume : ndarray or str
        3d volume to load.
        If string, it is used as a file path.
        If it is an ndarray, it is simply returned.

    Returns
    -------
    ndarray
    """
    viz_logger.info("Loading Volume...")
    if isinstance(volume, str):
        return nib.load(volume).get_fdata()
    else:
        return volume


class Viz:
    def __init__(self,
                 backend="fury"):
        """
        Set up visualization preferences.

        Parameters
        ----------
            backend : str, optional
                Should be either "fury" or "plotly".
                Default: "fury"
        """
        self.backend = backend
        if "fury" in backend:
            try:
                import AFQ.viz.fury_backend
            except ImportError:
                raise ImportError(viz_import_msg_error("fury"))
            self.visualize_bundles = AFQ.viz.fury_backend.visualize_bundles
            self.visualize_roi = AFQ.viz.fury_backend.visualize_roi
            self.visualize_volume = AFQ.viz.fury_backend.visualize_volume
            self.create_gif = AFQ.viz.fury_backend.create_gif
        elif "plotly" in backend:
            try:
                import AFQ.viz.plotly_backend
            except ImportError:
                raise ImportError(viz_import_msg_error("plotly"))
            self.visualize_bundles = AFQ.viz.plotly_backend.visualize_bundles
            self.visualize_roi = AFQ.viz.plotly_backend.visualize_roi
            self.visualize_volume = AFQ.viz.plotly_backend.visualize_volume
            self.create_gif = AFQ.viz.plotly_backend.create_gif
            self.single_bundle_viz = AFQ.viz.plotly_backend.single_bundle_viz
        else:
            raise TypeError("Visualization backend contain"
                            + " either 'plotly' or 'fury'. "
                            + "It is currently set to %s"
                            % backend)



import os.path as op
import random
import numpy as np
import numpy.testing as npt
import pytest

import nibabel as nib
import tempfile

from AFQ.models.csd import fit_csd
from AFQ.models.dti import fit_dti
from AFQ.tractography.tractography import track
from AFQ.utils.testing import make_tracking_data


seeds = np.array([[-80., -120., -60.],
                  [-81, -121, -61],
                  [-81, -120, -60]])


tmpdir = tempfile.TemporaryDirectory()
fbval = op.join(tmpdir.name, 'dti.bval')
fbvec = op.join(tmpdir.name, 'dti.bvec')
fdata = op.join(tmpdir.name, 'dti.nii.gz')
make_tracking_data(fbval, fbvec, fdata)

minlen = 20
step_size = 0.5


def test_csd_local_tracking():
    random.seed(1234)
    for sh_order in [4, 8, 10]:
        fname = fit_csd(fdata, fbval, fbvec,
                        response=((0.0015, 0.0003, 0.0003), 100),
                        sh_order=sh_order, lambda_=1, tau=0.1, mask=None,
                        out_dir=tmpdir.name)
        for directions in ["det", "prob"]:
            sls = track(
                fname,
                directions,
                odf_model="CSD",
                max_angle=30.,
                sphere=None,
                seed_mask=None,
                n_seeds=seeds,
                stop_mask=None,
                step_size=step_size,
                minlen=minlen,
                tracker="local").streamlines

            for sl in sls:
                npt.assert_(len(sl) >= minlen / step_size)


def test_dti_local_tracking():
    fdict = fit_dti(fdata, fbval, fbvec)
    for directions in ["det", "prob"]:
        sls = track(
            fdict['params'],
            directions,
            max_angle=30.,
            sphere=None,
            seed_mask=None,
            n_seeds=1,
            step_size=step_size,
            minlen=minlen,
            odf_model="DTI",
            tracker="local").streamlines
        for sl in sls:
            npt.assert_(len(sl) >= minlen / step_size)


def test_pft_tracking():
    for fname, odf in zip(
            [
                fit_dti(fdata, fbval, fbvec)['params'],
                fit_csd(
                    fdata, fbval, fbvec,
                    response=((0.0015, 0.0003, 0.0003), 100),
                    sh_order=8, lambda_=1, tau=0.1, mask=None,
                    out_dir=tmpdir.name)],
            ["DTI", "CSD"]):
        img = nib.load(fdata)
        data_shape  = img.shape
        data_affine = img.affine
        pve_wm_data = nib.Nifti1Image(np.ones(data_shape[:3]), img.affine)
        pve_gm_data = nib.Nifti1Image(np.zeros(data_shape[:3]), img.affine)
        pve_csf_data = nib.Nifti1Image(np.zeros(data_shape[:3]), img.affine)
        stop_mask = (pve_wm_data, pve_gm_data, pve_csf_data)

        for directions in ["det", "prob"]:
            for stop_threshold in ["ACT", "CMC"]:
                sls = track(
                    fname,
                    directions,
                    max_angle=30.,
                    sphere=None,
                    seed_mask=None,
                    stop_mask=stop_mask,
                    stop_threshold=stop_threshold,
                    n_seeds=1,
                    step_size=step_size,
                    minlen=minlen,
                    odf_model=odf,
                    tracker="pft").streamlines

                for sl in sls:
                    npt.assert_(len(sl) >= minlen / step_size)

    # Test error handling:
    with pytest.raises(RuntimeError):
        track(
            fname,
            directions,
            max_angle=30.,
            sphere=None,
            seed_mask=None,
            stop_mask=0,  # Stop mask needs to be a tuple!
            stop_threshold=stop_threshold,
            n_seeds=1,
            step_size=step_size,
            minlen=minlen,
            tracker="pft")

    with pytest.raises(RuntimeError):
        track(
            fname,
            directions,
            max_angle=30.,
            sphere=None,
            seed_mask=None,
            stop_mask=stop_mask,
            stop_threshold=None,  # Stop threshold needs to be a string!
            n_seeds=1,
            step_size=step_size,
            minlen=minlen,
            tracker="pft")


import numpy as np
import numpy.testing as npt
import AFQ.utils.stats as AFS


def test_contrast_index():
    x1 = np.asarray([1, 2, 3.0, 1.1])
    x2 = np.asarray([0.9, 3.0, 5.0, 0.8])
    contrast_index1 = AFS.contrast_index(x1, x2)

    npt.assert_almost_equal(contrast_index1, [
        0.05263158*2,
        -0.2*2,
        -0.25*2,
        0.15789474*2])



import logging
import numpy as np

import scipy.ndimage as ndim
from skimage.morphology import binary_dilation, convex_hull_image
from scipy.spatial.distance import dice

import nibabel as nib

from dipy.io.utils import (create_nifti_header, get_reference_info)
from dipy.tracking.streamline import select_random_set_of_streamlines
import dipy.tracking.utils as dtu

logger = logging.getLogger('AFQ')


def transform_inverse_roi(roi, mapping, bundle_name="ROI"):
    """
    After being non-linearly transformed, ROIs tend to have holes in them.
    We perform a couple of computational geometry operations on the ROI to
    fix that up.

    Parameters
    ----------
    roi : Nifti1Image, str, ndarray
        The ROI to transform. Can be a path or image, which will be
        converted to an ndarray.

    mapping : DiffeomorphicMap object
        A mapping between DWI space and a template.

    bundle_name : str, optional
        Name of bundle, which may be useful for error messages.
        Default: None

    Returns
    -------
    ROI after dilation and hole-filling
    """
    if isinstance(roi, str):
        roi = nib.load(roi)
    if isinstance(roi, nib.Nifti1Image):
        roi = roi.get_fdata()

    _roi = mapping.transform_inverse(roi, interpolation='linear')

    if np.sum(_roi) == 0:
        logger.warning(
            f'Lost ROI {bundle_name}, performing automatic binary dilation')
        _roi = binary_dilation(roi)
        _roi = mapping.transform_inverse(_roi, interpolation='linear')

    _roi = patch_up_roi(_roi > 0, bundle_name=bundle_name).astype(np.int32)

    return _roi


def patch_up_roi(roi, bundle_name="ROI"):
    """
    After being non-linearly transformed, ROIs tend to have holes in them.
    We perform a couple of computational geometry operations on the ROI to
    fix that up.

    Parameters
    ----------
    roi : 3D binary array
        The ROI after it has been transformed.

    bundle_name : str, optional
        Name of bundle, which may be useful for error messages.
        Default: None

    Returns
    -------
    ROI after dilation and hole-filling
    """
    hole_filled = ndim.binary_fill_holes(roi > 0)
    if not np.any(hole_filled):
        raise ValueError((
            f"{bundle_name} found to be empty after "
            "applying the mapping."))
    return hole_filled


def density_map(tractogram, n_sls=None, normalize=False):
    """
    Create a streamline density map.
    based on:
    https://dipy.org/documentation/1.1.1./examples_built/streamline_formats/

    Parameters
    ----------
    tractogram : StatefulTractogram
        Stateful tractogram whose streamlines are used to make
        the density map.
    n_sls : int or None, optional
        n_sls to randomly select to make the density map.
        If None, all streamlines are used.
        Default: None
    normalize : bool, optional
        Whether to normalize maximum values to 1.
        Default: False

    Returns
    -------
    Nifti1Image containing the density map.
    """
    tractogram.to_vox()

    sls = tractogram.streamlines
    if n_sls is not None:
        sls = select_random_set_of_streamlines(sls, n_sls)

    affine, vol_dims, voxel_sizes, voxel_order = get_reference_info(tractogram)
    tractogram_density = dtu.density_map(sls, np.eye(4), vol_dims)
    if normalize:
        tractogram_density = tractogram_density / tractogram_density.max()

    nifti_header = create_nifti_header(affine, vol_dims, voxel_sizes)
    density_map_img = nib.Nifti1Image(tractogram_density, affine, nifti_header)

    return density_map_img


def dice_coeff(arr1, arr2, weighted=True):
    """
    Compute Dice's coefficient between two images.

    Parameters
    ----------
    arr1 : Nifti1Image, str, ndarray
        One ndarray to compare. Can be a path or image, which will be
        converted to an ndarray.
    arr2 : Nifti1Image, str, ndarray
        The other ndarray to compare. Can be a path or image, which will be
        converted to an ndarray.
    weighted : bool, optional
        Whether or not to weight the DICE coefficient as in [Cousineau2017]_.
        The weighted Dice coefficient is calculated by adding the sum of all
        values in arr1 where arr2 is nonzero to the sum of all values in arr2
        where arr1 is nonzero, then dividing that by the sum of all values in
        arr1 and arr2.
        Default: True

    Returns
    -------
    The dice similarity between the images.

    Notes
    -----
    .. [1] Cousineau M, Jodoin PM, Morency FC, et al. A test-retest study on
           Parkinson's PPMI dataset yields statistically significant white
           matter fascicles. Neuroimage Clin. 2017;16:222-233. Published 2017
           Jul 25. doi:10.1016/j.nicl.2017.07.020
    """
    if isinstance(arr1, str):
        arr1 = nib.load(arr1)
    if isinstance(arr2, str):
        arr2 = nib.load(arr2)

    if isinstance(arr1, nib.Nifti1Image):
        arr1 = arr1.get_fdata()
    if isinstance(arr2, nib.Nifti1Image):
        arr2 = arr2.get_fdata()

    arr1 = arr1.flatten()
    arr2 = arr2.flatten()

    if weighted:
        return (
            np.sum(arr1 * arr2.astype(bool))
            + np.sum(arr2 * arr1.astype(bool)))\
            / (np.sum(arr1) + np.sum(arr2))
    else:
        # scipy's dice function returns the dice *dissimilarity*
        return 1 - dice(
            arr1.astype(bool),
            arr2.astype(bool))



from glob import glob
import shutil
import os
from sphinx_gallery.scrapers import figure_rst
class PNGScraper(object):
    def __init__(self):
        self.seen = set()

    def __repr__(self):
        return 'PNGScraper'

    def __call__(self, block, block_vars, gallery_conf):
        # Find all PNG files in the directory of this example.
        path_current_example = os.path.dirname(block_vars['src_file'])
        pngs = sorted(glob(os.path.join(path_current_example, '*.png')))

        # Iterate through PNGs, copy them to the sphinx-gallery output directory
        image_names = list()
        image_path_iterator = block_vars['image_path_iterator']
        for png in pngs:
            if png not in self.seen:
                self.seen |= set(png)
                this_image_path = image_path_iterator.next()
                image_names.append(this_image_path)
                shutil.move(png, this_image_path)
        # Use the `figure_rst` helper function to generate rST for image files
        return figure_rst(image_names, gallery_conf['src_dir'])



def contrast_index(x1, x2, double=True):
    """
    Calculate the contrast index between two arrays.
    The contrast index is symmetrical with respect to
    the choice of the baseline for comparison

    Parameters
    ----------
    x1 : ndarray of floats
        An ndarray to compare. The contrast index will have positive values
        where x1 > x2.
    x2 : ndarray of floats
        An ndarray to compare. The contrast index will have negative values
        where x2 > x1.
    double : bool, optional
        Double the contrast index.
        This makes it comparable to a percent difference.
        Default: True

    Returns
    -------
    contrast_index : ndarray of floats
        Contrast index calculated by doing (x1 - x2) / (x1 + x2)
    """
    ci = (x1 - x2) / (x1 + x2)
    if double:
        return ci * 2
    else:
        return ci



import cuslines

import numpy as np
from math import radians
from tqdm import tqdm
import logging

from dipy.data import small_sphere
from dipy.reconst.shm import OpdtModel, CsaOdfModel
from dipy.reconst import shm
from dipy.io.stateful_tractogram import StatefulTractogram, Space

from nibabel.streamlines.array_sequence import concatenate

from AFQ.tractography.utils import gen_seeds, get_percentile_threshold


logger = logging.getLogger('AFQ')


# Modified from https://github.com/dipy/GPUStreamlines/blob/master/run_dipy_gpu.py
def gpu_track(data, gtab, seed_img, stop_img, odf_model,
              seed_threshold, stop_threshold, thresholds_as_percentages,
              max_angle, step_size, n_seeds, random_seeds, rng_seed, ngpus,
              chunk_size):
    """
    Perform GPU tractography on DWI data.

    Parameters
    ----------
    data : ndarray
        DWI data.
    gtab : GradientTable
        The gradient table.
    seed_img : Nifti1Image
        Float or binary mask describing the ROI within which we seed for
        tracking.
    stop_img : Nifti1Image
        A float or binary mask that determines a stopping criterion
        (e.g. FA).
    odf_model : str, optional
        One of {"OPDT", "CSA"}
    seed_threshold : float
        The value of the seed_img above which tracking is seeded.
    stop_threshold : float
        The value of the stop_img below which tracking is
        terminated.
    thresholds_as_percentages : bool
        Interpret seed_threshold and stop_threshold as percentages of the
        total non-nan voxels in the seed and stop mask to include
        (between 0 and 100), instead of as a threshold on the
        values themselves. 
    max_angle : float
        The maximum turning angle in each step.
    step_size : float
        The size of a step (in mm) of tractography.
    n_seeds : int
        The seeding density: if this is an int, it is is how many seeds in each
        voxel on each dimension (for example, 2 => [2, 2, 2]). If this is a 2D
        array, these are the coordinates of the seeds. Unless random_seeds is
        set to True, in which case this is the total number of random seeds
        to generate within the mask. Default: 1
    random_seeds : bool
        If True, n_seeds is total number of random seeds to generate.
        If False, n_seeds encodes the density of seeds to generate.
    rng_seed : int
        random seed used to generate random seeds if random_seeds is
        set to True. Default: None    ngpus : int
        Number of GPUs to use.
    chunk_size : int
        Chunk size for GPU tracking.
    Returns
    -------
    """
    sh_order = 6

    seed_data = seed_img.get_fdata()
    stop_data = stop_img.get_fdata()

    if thresholds_as_percentages:
        stop_threshold = get_percentile_threshold(
            stop_data, stop_threshold)

    if odf_model.lower() == "opdt":
        model_type = cuslines.ModelType.OPDT
        model = OpdtModel(
            gtab,
            sh_order=sh_order,
            smooth=0.006,
            min_signal=1)
        fit_matrix = model._fit_matrix
        delta_b, delta_q = fit_matrix
    elif odf_model.lower() == "csa":
        model_type = cuslines.ModelType.CSAODF
        model = CsaOdfModel(
            gtab, sh_order=sh_order,
            smooth=0.006, min_signal=1)
        fit_matrix = model._fit_matrix
        delta_b = fit_matrix
        delta_q = fit_matrix
    else:
        raise ValueError((
            f"odf_model must be 'opdt' or "
            f"'csa', not {odf_model}"))

    sphere = small_sphere
    theta = sphere.theta
    phi = sphere.phi
    sampling_matrix, _, _ = shm.real_sym_sh_basis(sh_order, theta, phi)

    b0s_mask = gtab.b0s_mask
    dwi_mask = ~b0s_mask
    x, y, z = model.gtab.gradients[dwi_mask].T
    _, theta, phi = shm.cart2sphere(x, y, z)
    B, _, _ = shm.real_sym_sh_basis(sh_order, theta, phi)
    H = shm.hat(B)
    R = shm.lcr_matrix(H)

    sph_edges = sphere.edges
    sph_verticies = sphere.vertices

    gtargs = {}
    for var_name in [
        "data",
        "H", "R", "delta_b", "delta_q",
        "b0s_mask", "stop_data", "sampling_matrix",
            "sph_verticies", "sph_edges"]:
        var = locals()[var_name]

        if var_name in ["b0s_mask", "sph_edges"]:
            dtype = np.int32
        else:
            dtype = np.float64

        if not np.asarray(var).flags['C_CONTIGUOUS']:
            logger.warning(f"{var_name} is not C contiguous. Copying...")
            gtargs[var_name] = np.ascontiguousarray(
                var, dtype=dtype)
        else:
            gtargs[var_name] = np.asarray(var, dtype=dtype)

    gpu_tracker = cuslines.GPUTracker(
        model_type,
        radians(max_angle),
        1.0,
        stop_threshold,
        step_size,
        0.25,  # relative peak threshold
        radians(45),  # min separation angle
        gtargs["data"], gtargs["H"], gtargs["R"],
        gtargs["delta_b"], gtargs["delta_q"],
        gtargs["b0s_mask"], gtargs["stop_data"],
        gtargs["sampling_matrix"],
        gtargs["sph_verticies"], gtargs["sph_edges"],
        ngpus=ngpus, rng_seed=0)

    seeds = gen_seeds(
        seed_data, seed_threshold,
        n_seeds, thresholds_as_percentages,
        random_seeds, rng_seed, np.eye(4))

    global_chunk_sz = chunk_size * ngpus
    nchunks = (seeds.shape[0] + global_chunk_sz - 1) // global_chunk_sz

    streamlines_ls = [None] * nchunks
    with tqdm(total=seeds.shape[0]) as pbar:
        for idx in range(int(nchunks)):
            streamlines_ls[idx] = gpu_tracker.generate_streamlines(
                seeds[idx * global_chunk_sz:(idx + 1) * global_chunk_sz])
            pbar.update(
                seeds[idx * global_chunk_sz:(idx + 1) * global_chunk_sz].shape[0])

    sft = StatefulTractogram(
        concatenate(streamlines_ls, 0),
        seed_img, Space.VOX)

    return sft



"""
Registration tools
"""
import numpy as np
import nibabel as nib
from dipy.align.imwarp import DiffeomorphicMap

from dipy.align import syn_registration

import dipy.core.gradients as dpg
from dipy.align.streamlinear import whole_brain_slr


__all__ = ["syn_register_dwi", "write_mapping", "read_mapping",
           "slr_registration"]


def reduce_shape(shape):
    """
    Reduce dimension in shape to 3 if possible
    """
    try:
        return shape[:3]
    except TypeError:
        return shape


def syn_register_dwi(dwi, gtab, template=None, **syn_kwargs):
    """
    Register DWI data to a template.

    Parameters
    -----------
    dwi : nifti image or str
        Image containing DWI data, or full path to a nifti file with DWI.
    gtab : GradientTable or list of strings
        The gradients associated with the DWI data, or a string with [fbcal, ]
    template : nifti image or str, optional

    syn_kwargs : key-word arguments for :func:`syn_registration`

    Returns
    -------
    DiffeomorphicMap object
    """
    if template is None:
        import AFQ.data.fetch as afd
        template = afd.read_mni_template()
    if isinstance(template, str):
        template = nib.load(template)

    template_data = template.get_fdata()
    template_affine = template.affine

    if isinstance(dwi, str):
        dwi = nib.load(dwi)

    if not isinstance(gtab, dpg.GradientTable):
        gtab = dpg.gradient_table(*gtab)

    dwi_affine = dwi.affine
    dwi_data = dwi.get_fdata()
    mean_b0 = np.mean(dwi_data[..., gtab.b0s_mask], -1)
    warped_b0, mapping = syn_registration(mean_b0, template_data,
                                          moving_affine=dwi_affine,
                                          static_affine=template_affine,
                                          **syn_kwargs)
    return warped_b0, mapping


def write_mapping(mapping, fname):
    """
    Write out a syn registration mapping to file

    Parameters
    ----------
    mapping : a DiffeomorphicMap object derived from :func:`syn_registration`
    fname : str
        Full path to the nifti file storing the mapping

    """
    if isinstance(mapping, DiffeomorphicMap):
        mapping_imap = np.array([mapping.forward.T, mapping.backward.T]).T
        nib.save(nib.Nifti1Image(mapping_imap, mapping.codomain_world2grid),
                 fname)
    else:
        np.save(fname, mapping.affine)


def read_mapping(disp, domain_img, codomain_img, prealign=None):
    """
    Read a syn registration mapping from a nifti file

    Parameters
    ----------
    disp : str, Nifti1Image, or ndarray
        If string, file must of an image or ndarray.
        If image, contains the mapping displacement field in each voxel
        Shape (x, y, z, 3, 2)
        If ndarray, contains affine transformation used for mapping

    domain_img : str or Nifti1Image

    codomain_img : str or Nifti1Image

    Returns
    -------
    A :class:`DiffeomorphicMap` object
    """
    if isinstance(disp, str):
        if "nii.gz" in disp:
            disp = nib.load(disp)
        else:
            disp = np.load(disp)

    if isinstance(domain_img, str):
        domain_img = nib.load(domain_img)

    if isinstance(codomain_img, str):
        codomain_img = nib.load(codomain_img)

    if isinstance(disp, nib.Nifti1Image):
        mapping = DiffeomorphicMap(3, disp.shape[:3],
                                   disp_grid2world=np.linalg.inv(disp.affine),
                                   domain_shape=domain_img.shape[:3],
                                   domain_grid2world=domain_img.affine,
                                   codomain_shape=codomain_img.shape,
                                   codomain_grid2world=codomain_img.affine,
                                   prealign=prealign)

        disp_data = disp.get_fdata().astype(np.float32)
        mapping.forward = disp_data[..., 0]
        mapping.backward = disp_data[..., 1]
        mapping.is_inverse = True
    else:
        from AFQ.definitions.mapping import ConformedAffineMapping
        mapping = ConformedAffineMapping(
            disp,
            domain_grid_shape=reduce_shape(
                domain_img.shape),
            domain_grid2world=domain_img.affine,
            codomain_grid_shape=reduce_shape(
                codomain_img.shape),
            codomain_grid2world=codomain_img.affine)

    return mapping


def slr_registration(moving_data, static_data,
                     moving_affine=None, static_affine=None,
                     moving_shape=None, static_shape=None, **kwargs):
    """Register a source image (moving) to a target image (static).

    Parameters
    ----------
    moving : ndarray
        The source tractography data to be registered
    moving_affine : ndarray
        The affine associated with the moving (source) data.
    moving_shape : ndarray
        The shape of the space associated with the static (target) data.
    static : ndarray
        The target tractography data for registration
    static_affine : ndarray
        The affine associated with the static (target) data.
    static_shape : ndarray
        The shape of the space associated with the static (target) data.

    **kwargs:
        kwargs are passed into whole_brain_slr

    Returns
    -------
    AffineMap
    """
    from AFQ.definitions.mapping import ConformedAffineMapping

    _, transform, _, _ = whole_brain_slr(
        static_data, moving_data, x0='affine', verbose=False, **kwargs)

    return ConformedAffineMapping(
        transform,
        codomain_grid_shape=reduce_shape(static_shape),
        codomain_grid2world=static_affine,
        domain_grid_shape=reduce_shape(moving_shape),
        domain_grid2world=moving_affine)



import os.path as op
import logging

from AFQ.utils.path import drop_extension


logger = logging.getLogger('AFQ')


__all__ = ["Definition", "find_file", "name_from_path"]


class Definition(object):
    '''
    All Definitions should inherit this.
    For a given subject and session within the API, the definition is used
    to create a given image or map.
    Definitions have an init function which the users uses to specify
    how they want the definition to behave.
    The find_path function is called by the AFQ API.
    The api calls find_path to let the definition find relevant files
    for the given subject and session.
    '''

    def __init__(self):
        raise NotImplementedError("Please implement an __init__ method")

    def find_path(self, bids_layout, from_path,
                  subject, session, required=True):
        pass

    def str_for_toml(self):
        """
        Uses __init__ in str_for_toml to make string that will instantiate
        itself. Assumes object will have attributes of same name as
        __init__ args. This is important for reading/writing definitions
        as arguments to config files.
        """
        return type(self).__name__\
            + "("\
            + _arglist_to_string(
                self.__init__.__code__.co_varnames,
                get_attr=self)\
            + ')'


def _arglist_to_string(args, get_attr=None):
    '''
    Helper function
    Takes a list of arguments and unfolds them into a string.
    If get_attr is not None, it will be used to get the attribute
    corresponding to each argument instead.
    '''
    to_string = ""
    for arg in args:
        if arg == "self":
            continue
        if get_attr is not None:
            arg = getattr(get_attr, arg)
        if isinstance(arg, Definition):
            arg = arg.str_for_toml()
        elif isinstance(arg, str):
            arg = f"\"{arg}\""
        elif isinstance(arg, list):
            arg = f"[{_arglist_to_string(arg)}]"
        to_string = to_string + str(arg) + ', '
    if to_string[-2:] == ', ':
        to_string = to_string[:-2]
    return to_string


def name_from_path(path):
    file_name = op.basename(path)  # get file name
    file_name = drop_extension(file_name)  # remove extension
    if "-" in file_name:
        file_name = file_name.split("-")[-1]  # get suffix if exists
    return file_name


def _ff_helper(required, err_msg):
    if required:
        raise ValueError(err_msg)
    else:
        logger.warning(err_msg)
        return None


def find_file(bids_layout, path, filters, suffix, session, subject,
              extension=".nii.gz", required=True):
    """
    Helper function
    Generic calls to get_nearest to find a file
    """
    if "extension" not in filters:
        filters["extension"] = extension
    if "suffix" not in filters:
        filters["suffix"] = suffix

    # First, try to match the session.
    nearest = bids_layout.get_nearest(
        path,
        **filters,
        session=session,
        subject=subject,
        full_search=True,
        strict=False,
    )

    # If that fails, loosen session restriction
    # in order to find scans that are not session specific
    if nearest is None:
        nearest = bids_layout.get_nearest(
            path,
            **filters,
            subject=subject,
            full_search=True,
            strict=False,
        )

    # Nothing is found
    if nearest is None:
        return _ff_helper(required, (
            "No file found with these parameters:\n"
            f"suffix: {suffix},\n"
            f"session (searched with and without): {session},\n"
            f"subject: {subject},\n"
            f"filters: {filters},\n"
            f"near path: {path},\n"))

    path_subject = bids_layout.parse_file_entities(path).get(
        "subject", None
    )
    file_subject = bids_layout.parse_file_entities(nearest).get(
        "subject", None
    )
    path_session = bids_layout.parse_file_entities(path).get(
        "session", None
    )
    file_session = bids_layout.parse_file_entities(nearest).get(
        "session", None
    )

    # found file is wrong subject
    if path_subject != file_subject:
        return _ff_helper(required, (
            f"Expected subject IDs to match for the retrieved image file "
            f"and the supplied `from_path` file. Got sub-{file_subject} "
            f"from image file {nearest} and sub-{path_subject} "
            f"from `from_path` file {path}."))

    # found file is wrong session
    if (file_session is not None) and (path_session != file_session):
        return _ff_helper(required, (
            f"Expected session IDs to match for the retrieved image file "
            f"and the supplied `from_path` file. Got ses-{file_session} "
            f"from image file {nearest} and ses-{path_session} "
            f"from `from_path` file {path}."))

    return nearest



import os
import os.path as op

import numpy as np
import nibabel as nib

from dipy.reconst import fwdti

import AFQ.utils.models as ut


__all__ = ["fit_fwdti"]


def _fit(data, gtab, mask):
    fwmodel = fwdti.FreeWaterTensorModel(gtab)
    fwfit = fwmodel.fit(data, mask=mask)
    return fwfit


def fit_fwdti(data_files, bval_files, bvec_files, mask=None, out_dir=None,
              b0_threshold=50):
    """
    Fit the free water DTI model [1]_, save files with derived maps

    Parameters
    ----------
    data_files : str or list
        Files containing DWI data. If this is a str, that's the full path to a
        single file. If it's a list, each entry is a full path.
    bval_files : str or list
        Equivalent to `data_files`.
    bvec_files : str or list
        Equivalent to `data_files`.
    mask : ndarray, optional
        Binary mask, set to True or 1 in voxels to be processed.
        Default: Process all voxels.
    out_dir : str, optional
        A full path to a directory to store the maps that get computed.
        Default: maps get stored in the same directory as the last DWI file
        in `data_files`.
    b0_threshold : float


    Returns
    -------
    file_paths : a dict with the derived maps that were computed and full-paths
    to the files containing these maps.

    Note
    ----
    ..[1] R. Neto Henriques, A. Rokem, E. Garyfallidis, S. St-Jean, E.T.
          Peterson, M. Correia (2017). [Re] Optimization of a free water
          elimination two-compartment model for diffusion tensor imaging.
          *ReScience*

    """
    img, data, gtab, mask = ut.prepare_data(data_files, bval_files,
                                            bvec_files, mask=mask,
                                            b0_threshold=b0_threshold)

    fwfit = _fit(data, gtab, mask)

    FA = fwfit.fa
    MD = fwfit.md
    AD = fwfit.ad
    RD = fwfit.rd
    fwvf = fwfit.f
    params = fwfit.model_params

    maps = [FA, MD, AD, RD, fwvf, params]
    names = ['FA', 'MD', 'AD', 'RD', 'FWVF', 'params']

    if out_dir is None:
        if isinstance(data_files, list):
            out_dir = op.join(op.split(data_files[0])[0], 'fwdti')
        else:
            out_dir = op.join(op.split(data_files)[0], 'fwdti')

    if not op.exists(out_dir):
        os.makedirs(out_dir)

    aff = img.affine
    file_paths = {}
    for m, n in zip(maps, names):
        file_paths[n] = op.join(out_dir, 'fwdti_%s.nii.gz' % n)
        nib.save(nib.Nifti1Image(m, aff), file_paths[n])

    return file_paths



# This updates usage/config.rst to the latest cli
# Developers can run this after modifying any arguments the user can see

from AFQ.utils.bin import func_dict_to_arg_dict, dict_to_toml

prologue = """
The pyAFQ configuration file
----------------------------

This file should be a `toml <https://github.com/toml-lang/toml>`_ file. At
minimum, the file should contain the BIDS path::

    [files]
    bids_path = "path/to/study"


But additional configuration options can be provided.
See an example configuration file below::

    title = "My AFQ analysis"

"""

epilogue = \
    """
pyAFQ will store a copy of the configuration file alongside the computed
results. Note that the `title` variable and `[metadata]` section are both for
users to enter any title/metadata they would like and pyAFQ will generally
ignore them.
"""


def setup(app):
    arg_dict = func_dict_to_arg_dict()
    example_config = dict_to_toml(arg_dict)
    example_config = '    ' + example_config.replace('\n', '\n    ')

    with open('./source/reference/config.rst', 'w') as ff:
        ff.write(prologue + example_config + epilogue)



"""
================================
Delineating cerebellar peduncles
================================

The cerebellar peduncles are white matter tracts that connect the cerebellum to
the brainstem and cortex. In this example, we show how to delineate the
cerebellar peduncles using a subject from the Healthy Brain Network dataset.

This how-to will focus on the definition of the Cerebellar Peduncles (CP) based
on [1]_, [2].

"""

import AFQ.data.fetch as afd
import AFQ.api.bundle_dict as abd
from AFQ.api.group import GroupAFQ
from AFQ.definitions.image import RoiImage, ImageFile


"""
We will use a subject from the HBN dataset. When considering the data
for this operation, look to see whether the acquisition volume includes the
cerbellum. If it does not, it will be hard to delineate the CPs.
"""

bids_path = afd.fetch_hbn_afq(["NDARAA948VFH"])[1]

"""
The next line downloads the cerebellar peduncle templates from `Figshare <https://figshare.com/articles/dataset/Regions_of_interest_for_automated_fiber_quantification_of_cerebellar_bundles/23201630>`_.

"""

cp_rois = afd.read_cp_templates()


"""
The following line defines a bundle dictionary for the cerebellar
peduncles. There are three CPs: The ICP, the MCP, and the SCP. Each CP is
defined by two inclusion ROIs and one exclusion ROI. The Inferior CPs are
defined by inclusion ROIs. They do not decussate, so "cross_midline" is set to
False. The Superior CPs are defined by two inclusion ROIs and an exclusion ROI,
where each SCP's most superior inclusion ROI is the other SCP's exclusion ROI.
They decussate, so "cross_midline" is set to True. The Middle CPs are defined
by two inclusion ROIs and they use the SCP intermediate ROIs as exclusion ROIs.
"""

cp_bundles = abd.cerebellar_bd()

"""
The bundle dict has been defined, and now we are ready to run the AFQ pipeline.
In this case, we are using data that has been preprocessed with QSIprep, so
we have a brain mask that was generated from the T1w data of this subject.
"""

brain_mask_definition = ImageFile(
    suffix="mask",
    filters={'desc': 'brain',
             'space': 'T1w',
             'scope': 'qsiprep'})


"""
Next, we define a GroupAFQ object. In this case, the tracking parameters
focus specifically on the CP, by using the ``RoiImage`` class to define the
seed region. We seed extensively in the ROIs that define the CPs.
"""

cp_afq = GroupAFQ(
    name="cp_afq",
    bids_path=bids_path,
    preproc_pipeline="qsiprep",
    brain_mask_definition=brain_mask_definition,
    tracking_params={
        "n_seeds": 4,
        "directions": "prob",
        "odf_model": "CSD",
        "seed_mask": RoiImage()},
    clip_edges=True,
    bundle_info=cp_bundles)


"""
The call to `export("bundles")` triggers the execution of the full pipeline.
"""
cp_afq.export("bundles")

"""

References
----------
.. [1] S. Jossinger, A. Sares, A. Zislis, D. Sury, V. Gracco, M. Ben-Shachar (2022)
       White matter correlates of sensorimotor synchronization in persistent
       developmental stuttering, Journal of Communication Disorders, 95.

.. [2] S. Jossinger, M. Yablonski, O. Amir, M. Ben-Shachar (2023). The
       contributions of the cerebellar peduncles and the frontal aslant tract
       in mediating speech fluency. Neurobiology of Language 2023;
       doi: https://doi.org/10.1162/nol_a_00098

"""



BUNDLE_RECO_2_AFQ = \
    {
        "AF_L": "Left Arcuate", "AF_R": "Right Arcuate",
        "UF_L": "Left Uncinate", "UF_R": "Right Uncinate",
        "IFOF_L": "Left Inferior Fronto-occipital",
        "IFOF_R": "Right Inferior Fronto-occipital",
        "CST_L": "Left Corticospinal", "CST_R": "Right Corticospinal",
        "ILF_L": "Left Inferior Longitudinal",
        "ILF_R": "Right Inferior Longitudinal",
        "SLF_L": "Left Superior Longitudinal",
        "SLF_R": "Right Superior Longitudinal"
    }

BUNDLE_MAT_2_PYTHON = {
    'RightCorticospinal': 'Right Corticospinal',
    'LeftCorticospinal': 'Left Corticospinal',
    'RightUncinate': 'Right Uncinate', 'LeftUncinate': 'Left Uncinate',
    'Left IFOF': 'Left Inferior Fronto-occipital',
    'Right IFOF': 'Right Inferior Fronto-occipital',
    'LeftIFOF': 'Left Inferior Fronto-occipital',
    'RightIFOF': 'Right Inferior Fronto-occipital',
    'RightArcuate': 'Right Arcuate', 'LeftArcuate': 'Left Arcuate',
    'Right Thalamic Radiation': 'Right Anterior Thalamic',
    'Left Thalamic Radiation': 'Left Anterior Thalamic',
    'RightThalamicRadiation': 'Right Anterior Thalamic',
    'LeftThalamicRadiation': 'Left Anterior Thalamic',
    'Right Cingulum Cingulate': 'Right Cingulum Cingulate',
    'Left Cingulum Cingulate': 'Left Cingulum Cingulate',
    'RightCingulumCingulate': 'Right Cingulum Cingulate',
    'LeftCingulumCingulate': 'Left Cingulum Cingulate',
    'Callosum Forceps Major': 'Forceps Major',
    'Callosum Forceps Minor': 'Forceps Minor',
    'CallosumForcepsMajor': 'Forceps Major',
    'CallosumForcepsMinor': 'Forceps Minor',
    'Right ILF': 'Right Inferior Longitudinal',
    'Left ILF': 'Left Inferior Longitudinal',
    'RightILF': 'Right Inferior Longitudinal',
    'LeftILF': 'Left Inferior Longitudinal',
    'Right SLF': 'Right Superior Longitudinal',
    'Left SLF': 'Left Superior Longitudinal',
    'RightSLF': 'Right Superior Longitudinal',
    'LeftSLF': 'Left Superior Longitudinal'}



# -*- coding: utf-8 -*-# -*- coding: utf-8 -*-
import warnings
import tempfile

from AFQ.definitions.mapping import SynMap
from AFQ.definitions.utils import Definition
import AFQ.api.bundle_dict as abd
warnings.simplefilter(action='ignore', category=FutureWarning)  # noqa

import logging
from AFQ.api.participant import ParticipantAFQ
from AFQ.api.utils import (
    check_attribute, AFQclass_doc,
    export_all_helper, valid_exports_string)
import AFQ.utils.streamlines as aus
from AFQ.viz.utils import get_eye

from dipy.utils.parallel import paramap
from dipy.io.stateful_tractogram import StatefulTractogram, Space
import dipy.tracking.streamlinespeed as dps
import dipy.tracking.streamline as dts
from dipy.io.streamline import save_tractogram

from AFQ.version import version as pyafq_version
from AFQ.viz.utils import trim
import pandas as pd
import pydra
import numpy as np
import os
import os.path as op
from tqdm import tqdm
import json
import s3fs
from time import time
import nibabel as nib
from PIL import Image
from s3bids.utils import S3BIDSStudy
import glob

from bids.layout import BIDSLayout, BIDSLayoutIndexer
try:
    import afqbrowser as afqb
    using_afqb = True
except ImportError:
    using_afqb = False


__all__ = ["GroupAFQ", "get_afq_bids_entities_fname"]


logger = logging.getLogger('AFQ')
logger.setLevel(logging.INFO)


# get rid of unnecessary columns in df
def clean_pandas_df(df):
    df = df.reset_index(drop=True)
    df = df.loc[:, ~df.columns.str.contains('^Unnamed')]
    return df


def get_afq_bids_entities_fname():
    return op.dirname(op.dirname(op.abspath(
        aus.__file__))) + "/afq_bids_entities.json"


class _ParticipantAFQInputs:
    def __init__(
            self, dwi_data_file, bval_file, bvec_file, results_dir,
            kwargs):
        self.dwi_data_file = dwi_data_file
        self.bval_file = bval_file
        self.bvec_file = bvec_file
        self.results_dir = results_dir
        self.kwargs = kwargs


class GroupAFQ(object):
    f"""{AFQclass_doc}"""

    def __init__(self,
                 bids_path,
                 bids_filters={"suffix": "dwi"},
                 preproc_pipeline="all",
                 participant_labels=None,
                 output_dir=None,
                 parallel_params={"engine": "serial"},
                 bids_layout_kwargs={},
                 **kwargs):
        '''
        Initialize a GroupAFQ object from a BIDS dataset.

        Parameters
        ----------
        bids_path : str
            The path to preprocessed diffusion data organized in a BIDS
            dataset. This should contain a BIDS derivative dataset with
            preprocessed dwi/bvals/bvecs.
        bids_filters : dict
            Filter to pass to bids_layout.get when finding DWI files.
            Default: {"suffix": "dwi"}
        preproc_pipeline : str, optional.
            The name of the pipeline used to preprocess the DWI data.
            Default: "all".
        participant_labels : list or None, optional
            List of participant labels (subject IDs) to perform
            processing on. If None, all subjects are used.
            Default: None
        output_dir : str or None, optional
            Path to output directory. If None, outputs are put
            in a AFQ pipeline folder in the derivatives folder of
            the BIDS directory. pyAFQ will use existing derivatives
            from the output directory if they exist, instead of recalculating
            them (this means you need to clear the output folder if you want
            to recalculate a derivative).
            Default: None
        parallel_params : dict, optional
            Parameters to pass to paramap in AFQ.utils.parallel,
            to parallelize computations across subjects and sessions.
            Set "n_jobs" to -1 to automatically parallelize as
            the number of cpus. Here is an example for how to do
            multiprocessing with 4 cpus:
            {"n_jobs": 4, "engine": "joblib", "backend": "loky"}
            Default: {"engine": "serial"}
        bids_layout_kwargs: dict, optional
            Additional arguments to give to BIDSLayout from pybids.
            For large datasets, try:
            {"validate": False, "index_metadata": False}
            Default: {}
        kwargs : additional optional parameters
            You can set additional parameters for any step
            of the process. See :ref:`usage/kwargs` for more details.

        Examples
        --------
        api.GroupAFQ(my_path, csd_sh_order=4)
        api.GroupAFQ(
            my_path,
            reg_template_spec="mni_t2", reg_subject_spec="b0")
        '''
        if not isinstance(bids_path, str):
            raise TypeError("bids_path must be a string")
        if not op.exists(bids_path):
            raise ValueError("bids_path not found")
        if not op.exists(op.join(bids_path, "dataset_description.json")):
            raise ValueError("There must be a dataset_description.json"
                             + " in bids_path")
        if not isinstance(bids_filters, dict):
            raise TypeError("bids_filters must be a dict")
        # preproc_pipeline typechecking handled by pyBIDS
        if participant_labels is not None\
                and not isinstance(participant_labels, list):
            raise TypeError(
                "participant_labels must be either a list or None")
        if output_dir is not None\
                and not isinstance(output_dir, str):
            raise TypeError(
                "output_dir must be either a str or None")
        if not isinstance(parallel_params, dict):
            raise TypeError("parallel_params must be a dict")
        if not isinstance(bids_layout_kwargs, dict):
            raise TypeError("bids_layout_kwargs must be a dict")

        parallel_params["engine"] = parallel_params.get("engine", "serial")
        self.logger = logger

        self.parallel_params = parallel_params
        self.wf_dict = {}

        # validate input and fail early
        if not op.exists(bids_path):
            raise ValueError(f'Unable to locate BIDS dataset in: {bids_path}')

        # This is where all the outputs will go:
        if output_dir is None:
            self.afq_path = op.join(bids_path, 'derivatives', 'afq')
            self.afqb_path = op.join(bids_path, 'derivatives', 'afq_browser')
        else:
            self.afq_path = output_dir
            self.afqb_path = op.join(output_dir, 'afq_browser')

        # Create it as needed:
        os.makedirs(self.afq_path, exist_ok=True)

        bids_indexer = BIDSLayoutIndexer(**bids_layout_kwargs)
        bids_layout = BIDSLayout(
            bids_path, derivatives=True, indexer=bids_indexer)
        bids_description = bids_layout.description

        # check that any files exist in the derivatives folder,
        # not including the dataset_description.json files
        # the second check may be particularly useful in checking
        # that the derivatives folder is well-defined
        if len(bids_layout.get())\
                - len(bids_layout.get(extension="json")) < 1:
            raise ValueError(
                f"No non-json files recognized by pyBIDS in {bids_path}")
        if len(bids_layout.get(scope=preproc_pipeline))\
                - len(bids_layout.get(
                    scope=preproc_pipeline,
                    extension="json")) < 1:
            raise ValueError((
                f"No non-json files recognized by "
                f"pyBIDS in the pipeline: {preproc_pipeline}"))

        # Add required metadata file at top level (inheriting as needed):
        pipeline_description = {
            "Name": bids_description["Name"],
            "BIDSVersion": bids_description["BIDSVersion"],
            "PipelineDescription": {"Name": "pyAFQ",
                                    "Version": pyafq_version},
            "GeneratedBy": [{"Name": op.basename(self.afq_path),
                             "Version": pyafq_version}]}

        pl_desc_file = op.join(self.afq_path, 'dataset_description.json')

        with open(pl_desc_file, 'w') as outfile:
            json.dump(pipeline_description, outfile)

        self.subjects = bids_layout.get(return_type='id', target='subject')
        if not len(self.subjects):
            raise ValueError(
                "`bids_path` contains no subjects in derivatives folders."
                + " This could be caused by derivatives folders not following"
                + " the BIDS format.")

        if participant_labels is not None:
            filtered_subjects = []
            subjects_found_printed = False
            for subjectID in participant_labels:
                subjectID = str(subjectID)
                if subjectID not in self.subjects:
                    self.logger.warning((
                        f"Subject {subjectID} specified in "
                        f"`participant_labels` but not found "
                        f"in BIDS derivatives folders"))
                    if not subjects_found_printed:
                        subjects_found_printed = True
                        self.logger.warning((
                            f"Only these subjects found in BIDS "
                            f"derivatives folders: {self.subjects}"))
                else:
                    filtered_subjects.append(subjectID)
            self.subjects = filtered_subjects
            if not len(self.subjects):
                raise ValueError(
                    "No subjects specified in `participant_labels` "
                    + " found in BIDS derivatives folders."
                    + " See above warnings.")

        sessions = bids_layout.get(return_type='id', target='session')
        self.sessions = sessions if len(sessions) else [None]

        # do not bother to parallelize if less than 2 subject-sessions
        if len(self.sessions) * len(self.subjects) < 2:
            self.parallel_params["engine"] = "serial"

        # do not parallelize segmentation if parallelizing across
        # subject-sessions
        if self.parallel_params["engine"] != "serial":
            if "segmentation_params" not in kwargs:
                kwargs["segmentation_params"] = {}
            if "parallel_segmentation" not in kwargs["segmentation_params"]:
                kwargs["segmentation_params"]["parallel_segmentation"] = {}
            kwargs["segmentation_params"]["parallel_segmentation"]["engine"] =\
                "serial"

        self.valid_sub_list = []
        self.valid_ses_list = []
        self.pAFQ_list = []
        self.pAFQ_inputs_list = []
        for subject in self.subjects:
            self.wf_dict[subject] = {}
            for session in self.sessions:
                this_kwargs = kwargs.copy()
                results_dir = op.join(self.afq_path, 'sub-' + subject)

                if session is not None:
                    results_dir = op.join(results_dir, 'ses-' + session)

                dwi_bids_filters = {
                    "subject": subject,
                    "session": session,
                    "return_type": "filename",
                    "scope": preproc_pipeline,
                    "extension": "nii.gz",
                    "suffix": "dwi",
                }
                dwi_bids_filters.update(bids_filters)
                dwi_files = bids_layout.get(**dwi_bids_filters)

                if (not len(dwi_files)):
                    self.logger.warning(
                        f"No dwi found for subject {subject} and session "
                        f"{session}. Skipping.")
                    continue

                os.makedirs(results_dir, exist_ok=True)
                dwi_data_file = dwi_files[0]

                # For bvals and bvecs, use ``get_bval()`` and ``get_bvec()`` to
                # walk up the file tree and inherit the closest bval and bvec
                # files. Maintain input ``bids_filters`` in case user wants to
                # specify acquisition labels, but pop suffix since it is
                # already specified inside ``get_bvec()`` and ``get_bval()``
                suffix = bids_filters.pop("suffix", None)
                bvec_file = bids_layout.get_bvec(
                    dwi_data_file,
                    **bids_filters)
                bval_file = bids_layout.get_bval(
                    dwi_data_file,
                    **bids_filters)
                if suffix is not None:
                    bids_filters["suffix"] = suffix

                # Call find path for all definitions
                for key, value in this_kwargs.items():
                    if key == "scalars":
                        for scalar in this_kwargs["scalars"]:
                            if isinstance(scalar, Definition):
                                scalar_found = scalar.find_path(
                                    bids_layout,
                                    dwi_data_file,
                                    subject,
                                    session,
                                    required=False)
                                if scalar_found is False:
                                    this_kwargs["scalars"].remove(scalar)
                    elif key == "import_tract":
                        if isinstance(this_kwargs["import_tract"], dict):
                            it_res = \
                                bids_layout.get(
                                    subject=subject,
                                    session=session,
                                    extension=[
                                        '.trk',
                                        '.tck',
                                        '.vtk',
                                        '.fib',
                                        '.dpy'],
                                    return_type='filename',
                                    **this_kwargs["import_tract"])
                            if len(it_res) < 1:
                                raise ValueError((
                                    "No custom tractography found for"
                                    f" subject {subject}"
                                    " and session "
                                    f"{session}."))
                            elif len(it_res) > 1:
                                this_kwargs["import_tract"] = it_res[0]
                                logger.warning((
                                    f"Multiple viable custom tractographies found for"
                                    f" subject "
                                    f"{subject} and session "
                                    f"{session}. Will use: {it_res[0]}"))
                            else:
                                this_kwargs["import_tract"] = it_res[0]
                    elif isinstance(value, dict):
                        for _, subvalue in value.items():
                            if isinstance(subvalue, Definition):
                                subvalue.find_path(
                                    bids_layout,
                                    dwi_data_file,
                                    subject,
                                    session)
                    elif isinstance(value, Definition):
                        value.find_path(
                            bids_layout,
                            dwi_data_file,
                            subject,
                            session)

                # call find path for all ROIs
                if "bundle_info" in this_kwargs and isinstance(
                        this_kwargs["bundle_info"], abd.BundleDict):
                    for b_name in this_kwargs["bundle_info"].bundle_names:
                        this_kwargs["bundle_info"].apply_to_rois(
                            b_name,
                            this_kwargs["bundle_info"]._use_bids_info,
                            bids_layout, dwi_data_file, subject, session,
                            dry_run=False)

                self.valid_sub_list.append(subject)
                self.valid_ses_list.append(str(session))

                this_pAFQ_inputs = _ParticipantAFQInputs(
                    dwi_data_file,
                    bval_file, bvec_file,
                    results_dir,
                    this_kwargs)
                this_pAFQ = ParticipantAFQ(
                    this_pAFQ_inputs.dwi_data_file,
                    this_pAFQ_inputs.bval_file,
                    this_pAFQ_inputs.bvec_file,
                    this_pAFQ_inputs.results_dir,
                    **this_pAFQ_inputs.kwargs)
                self.wf_dict[subject][str(session)] = this_pAFQ.wf_dict
                self.pAFQ_list.append(this_pAFQ)
                self.pAFQ_inputs_list.append(this_pAFQ_inputs)

    def combine_profiles(self):
        tract_profiles_dict = self.export("profiles")
        if len(self.sessions) > 1:
            tract_profiles_list = []
            for _, subject_dict in tract_profiles_dict.items():
                tract_profiles_list.extend(subject_dict.values())
        else:
            tract_profiles_list = list(tract_profiles_dict.values())
        _df = combine_list_of_profiles(tract_profiles_list)
        out_file = op.abspath(op.join(
            self.afq_path, "tract_profiles.csv"))
        os.makedirs(op.dirname(out_file), exist_ok=True)
        _df = clean_pandas_df(_df)
        _df.to_csv(out_file, index=False)
        return _df

    def get_streamlines_json(self):
        sls_json_fname = op.abspath(op.join(
            self.afq_path, "afqb_streamlines.json"))
        if not op.exists(sls_json_fname):
            subses_info = []

            def load_next_subject():
                subses_idx = len(subses_info)
                sub = self.valid_sub_list[subses_idx]
                ses = self.valid_ses_list[subses_idx]
                this_bundles_file = self.export(
                    "bundles", collapse=False)[sub][ses]
                this_mapping = self.export("mapping", collapse=False)[sub][ses]
                this_img = self.export(
                    "dwi", collapse=False)[sub][ses]
                seg_sft = aus.SegmentedSFT.fromfile(
                    this_bundles_file,
                    this_img)
                seg_sft.sft.to_rasmm()
                subses_info.append((seg_sft, this_mapping))

            bundle_dict = self.export("bundle_dict", collapse=False)[
                self.valid_sub_list[0]][self.valid_ses_list[0]]

            sls_dict = {}
            load_next_subject()  # load first subject
            for b in bundle_dict.keys():
                for i in range(len(self.valid_sub_list)):
                    seg_sft, mapping = subses_info[i]
                    idx = seg_sft.bundle_idxs[b]
                    # use the first subses that works
                    # otherwise try each successive subses
                    if len(idx) == 0:
                        # break if we run out of subses
                        if i + 1 >= len(self.valid_sub_list):
                            break
                        # load subses if not already loaded
                        if i + 1 >= len(subses_info):
                            load_next_subject()
                        continue
                    if len(idx) > 100:
                        idx = np.random.choice(
                            idx, size=100, replace=False)
                    these_sls = seg_sft.sft.streamlines[idx]
                    these_sls = dps.set_number_of_points(these_sls, 100)
                    tg = StatefulTractogram(
                        these_sls,
                        seg_sft.sft,
                        Space.RASMM)
                    delta = dts.values_from_volume(
                        mapping.forward,
                        tg.streamlines, np.eye(4))
                    moved_sl = dts.Streamlines(
                        [d + s for d, s in zip(delta, tg.streamlines)])
                    moved_sl = np.asarray(moved_sl)
                    median_sl = np.median(moved_sl, axis=0)
                    sls_dict[b] = {"coreFiber": median_sl.tolist()}
                    for ii, sl_idx in enumerate(idx):
                        sls_dict[b][str(sl_idx)] = moved_sl[ii].tolist()
                    break

            with open(sls_json_fname, 'w') as fp:
                json.dump(sls_dict, fp)
        return sls_json_fname

    def export(self, attr_name="help", collapse=True):
        f"""
        Export a specific output. To print a list of available outputs,
        call export without arguments.
        {valid_exports_string}

        Parameters
        ----------
        attr_name : str
            Name of the output to export. Default: "help"
        collapse : bool
            Whether to collapse session dimension if there is only 1 session.
            Default: True

        Returns
        -------
        output : dict
            The specific output as a dictionary. Keys are subjects.
            Values are dictionaries with keys of sessions
            if multiple sessions are used. Otherwise, values are
            the output.
            None if called without arguments.
        """
        section = check_attribute(attr_name)

        # iterate over subjects / sessions,
        # decide if they need to be calculated or not
        in_list = []
        to_calc_list = []
        results = {}
        for ii, subject in enumerate(self.valid_sub_list):
            if subject not in results:
                results[subject] = {}
            session = self.valid_ses_list[ii]
            wf_dict = self.wf_dict[subject][str(session)]
            if section is not None:
                wf_dict = wf_dict[section]
            if ((self.parallel_params.get("engine", False) != "serial")
                    and (hasattr(wf_dict, "efferents"))
                    and (attr_name not in wf_dict.efferents)):
                in_list.append((wf_dict))
                to_calc_list.append((subject, session))
            else:
                results[subject][session] = wf_dict[attr_name]

        # if some need to be calculated, do those in parallel
        if to_calc_list:
            par_results = paramap(
                lambda wf, attr: wf[attr], in_list,
                func_args=[attr_name],
                **self.parallel_params)

            for i, subses in enumerate(to_calc_list):
                subject, session = subses
                results[subject][session] = par_results[i]

        # If only one session, collapse session dimension
        if len(self.sessions) == 1 and collapse:
            for subject in self.valid_sub_list:
                results[subject] = results[subject][self.valid_ses_list[0]]

        return results

    def export_up_to(self, attr_name="help"):
        f"""
        Export all derivatives necessary for a specific output.
        To print a list of available outputs,
        call export_up_to without arguments.
        {valid_exports_string}

        Parameters
        ----------
        attr_name : str
            Name of the output to export up to. Default: "help"
        """
        section = check_attribute(attr_name)
        wf_dict = self.wf_dict[
            self.valid_sub_list[0]][self.valid_ses_list[0]]
        if section is not None:
            wf_dict = wf_dict[section]
        for dependent in wf_dict.plan.dependencies[attr_name]:
            self.export(dependent)

    def export_all(self, viz=True, afqbrowser=True, xforms=True,
                   indiv=True):
        """ Exports all the possible outputs

        Parameters
        ----------
        viz : bool
            Whether to output visualizations. This includes tract profile
            plots, a figure containing all bundles, and, if using the AFQ
            segmentation algorithm, individual bundle figures.
            Default: True
        afqbrowser : bool
            Whether to output an AFQ-Browser from this AFQ instance.
            Default: True
        xforms : bool
            Whether to output the reg_template image in subject space and,
            depending on if it is possible based on the mapping used, to
            output the b0 in template space.
            Default: True
        indiv : bool
            Whether to output individual bundles in their own files, in
            addition to the one file containing all bundles. If using
            the AFQ segmentation algorithm, individual ROIs are also
            output.
            Default: True
        """
        start_time = time()

        export_all_helper(self, xforms, indiv, viz)

        self.combine_profiles()
        if afqbrowser:
            self.assemble_AFQ_browser()
        self.logger.info(
            f"Time taken for export all: {str(time() - start_time)}")

    def cmd_outputs(self, cmd="rm", dependent_on=None, exceptions=[],
                    suffix=""):
        """
        Perform some command some or all outputs of pyafq.
        This is useful if you change a parameter and need
        to recalculate derivatives that depend on it.
        Some examples: cp, mv, rm .
        -r will be automtically added when necessary.

        Parameters
        ----------
        cmd : str
            Command to run on outputs. Default: 'rm'
        dependent_on : str or None
            Which derivatives to perform command on .
            If None, perform on all.
            If "track", perform on all derivatives that depend on the
            tractography.
            If "recog", perform on all derivatives that depend on the
            bundle recognition.
            Default: None
        exceptions : list of str
            Name outputs that the command should not be applied to.
            Default: []
        suffix : str
            Parts of command that are used after the filename.
            Default: ""

        Example
        -------
        # This command would move all derivatives that are
        # dependent on the tractography into 'my_other_folder'
        myafq.cmd_outputs(
            "cp",
            dependent_on="track",
            suffix="~/my_other_folder/")
        """
        for pAFQ in self.pAFQ_list:
            pAFQ.cmd_outputs(cmd, dependent_on, exceptions, suffix=suffix)

    clobber = cmd_outputs  # alias for default of cmd_outputs

    def make_all_participant_montages(self, images_per_row=2):
        """
        Generate montage of all bundles for a all subjects.

        Parameters
        ----------
        images_per_row : int
            Number of bundle images per row in output file.
            Default: 2

        Returns
        -------
        filename of montage images
        """
        for pAFQ in self.pAFQ_list:
            pAFQ.participant_montage(images_per_row=images_per_row)

    def group_montage(self, bundle_name, size, view, direc, slice_pos=None):
        """
        Generate montage file(s) of a given bundle at a given angle.

        Parameters
        ----------
        bundle_name : str
            Name of bundle to visualize, should be the same as in the
            bundle dictionary.
        size : tuple of int
            The number of columns and rows for each file.
        view : str
            Which view to display. Can be one of sagittal, coronal, or axial.
        direc : str
            Which direction to views. Can be one of left, right, top, bottom, 
            front, back
        slice_pos : float, or None
            If float, indicates the fractional position along the
            perpendicular axis to the slice. Currently only works with plotly.
            If None, no slice is displayed.

        Returns
        -------
        list of filenames of montage images
        """
        tdir = tempfile.gettempdir()

        best_scalar = self.export("best_scalar", collapse=False)[
            self.valid_sub_list[0]][self.valid_ses_list[0]]

        viz_backend_dict = self.export("viz_backend", collapse=False)
        b0_backend_dict = self.export("b0", collapse=False)
        dwi_affine_dict = self.export("dwi_affine", collapse=False)
        bundles_dict = self.export("bundles", collapse=False)
        best_scalar_dict = self.export(best_scalar, collapse=False)

        all_fnames = []
        self.logger.info("Generating Montage...")
        for ii in tqdm(range(len(self.valid_ses_list))):
            this_sub = self.valid_sub_list[ii]
            this_ses = self.valid_ses_list[ii]
            viz_backend = viz_backend_dict[this_sub][this_ses]
            b0 = b0_backend_dict[this_sub][this_ses]
            dwi_affine = dwi_affine_dict[this_sub][this_ses]
            bundles = bundles_dict[this_sub][this_ses]
            best_scalar = best_scalar_dict[this_sub][this_ses]

            flip_axes = [False, False, False]
            for i in range(3):
                flip_axes[i] = (dwi_affine[i, i] < 0)

            if slice_pos is not None:
                slice_kwargs = {}
                if view == "sagittal":
                    slice_kwargs["x_pos"] = slice_pos
                    slice_kwargs["y_pos"] = None
                    slice_kwargs["z_pos"] = None
                elif view == "coronal":
                    slice_kwargs["x_pos"] = None
                    slice_kwargs["y_pos"] = slice_pos
                    slice_kwargs["z_pos"] = None
                elif view == "axial":
                    slice_kwargs["x_pos"] = None
                    slice_kwargs["y_pos"] = None
                    slice_kwargs["z_pos"] = slice_pos

                figure = viz_backend.visualize_volume(
                    best_scalar,
                    opacity=1.0,
                    flip_axes=flip_axes,
                    interact=False,
                    inline=False,
                    **slice_kwargs)
            else:
                figure = None

            figure = viz_backend.visualize_bundles(
                bundles,
                shade_by_volume=best_scalar,
                flip_axes=flip_axes,
                bundle=bundle_name,
                interact=False,
                inline=False,
                figure=figure)

            eye = get_eye(view, direc)

            this_fname = tdir + f"/t{ii}.png"
            if "plotly" in viz_backend.backend:

                figure.update_layout(scene_camera=dict(
                    projection=dict(type="orthographic"),
                    up={"x": 0, "y": 0, "z": 1},
                    eye=eye,
                    center=dict(x=0, y=0, z=0)))
                figure.write_image(this_fname)

                # temporary fix for memory leak
                import plotly.io as pio
                pio.kaleido.scope._shutdown_kaleido()
            else:
                from dipy.viz import window
                direc = np.fromiter(eye.values(), dtype=int)
                data_shape = np.asarray(nib.load(b0).get_fdata().shape)
                figure.set_camera(
                    position=direc * data_shape,
                    focal_point=data_shape // 2,
                    view_up=(0, 0, 1))
                figure.zoom(0.5)
                window.snapshot(figure, fname=this_fname, size=(600, 600))

        def _save_file(curr_img, curr_file_num):
            save_path = op.abspath(op.join(
                self.afq_path,
                (f"bundle-{bundle_name}_view-{view}"
                    f"_idx-{curr_file_num}_montage.png")))
            curr_img.save(save_path)
            all_fnames.append(save_path)

        this_img_trimmed = {}
        max_height = 0
        max_width = 0
        for ii in range(len(self.valid_ses_list)):
            this_img = Image.open(tdir + f"/t{ii}.png")
            try:
                this_img_trimmed[ii] = trim(trim(this_img))
            except IndexError:  # this_img is a picture of nothing
                this_img_trimmed[ii] = this_img

            if this_img_trimmed[ii].size[0] > max_width:
                max_width = this_img_trimmed[ii].size[0]
            if this_img_trimmed[ii].size[1] > max_height:
                max_height = this_img_trimmed[ii].size[1]

        curr_img = Image.new(
            'RGB',
            (max_width * size[0], max_height * size[1]),
            color="white")
        curr_file_num = 0
        for ii in range(len(self.valid_ses_list)):
            x_pos = ii % size[0]
            _ii = ii // size[0]
            y_pos = _ii % size[1]
            _ii = _ii // size[1]
            file_num = _ii

            if file_num != curr_file_num:
                _save_file(curr_img, curr_file_num)
                curr_img = Image.new(
                    'RGB',
                    (max_width * size[0], max_height * size[1]),
                    color="white")
                curr_file_num = file_num
            curr_img.paste(
                this_img_trimmed[ii],
                (x_pos * max_width, y_pos * max_height))

        _save_file(curr_img, curr_file_num)
        return all_fnames

    def combine_bundle(self, bundle_name):
        """
        Transforms a given bundle to reg_template space for all subjects
        then merges them to one trk file.
        Useful for visualizing the variability in the bundle across subjects.
        Note: currently only implemented using built-in SynMap

        Parameters
        ----------
        bundle_name : str
        Name of the bundle to transform, should be one of the bundles in
        bundle_dict.
        """
        reference_wf_dict = self.wf_dict[
            self.valid_sub_list[0]][self.valid_ses_list[0]]
        if "mapping_definition" in reference_wf_dict:
            mapping_definition = reference_wf_dict["mapping_definition"]
            if mapping_definition is not None and not isinstance(
                    mapping_definition, SynMap):
                raise NotImplementedError((
                    "combine_bundle not implemented for mapping_definition "
                    "other than SynMap"))

        reg_template = self.export("reg_template", collapse=False)[
            self.valid_sub_list[0]][self.valid_ses_list[0]]
        bundles_dict = self.export("bundles", collapse=False)
        mapping_dict = self.export("mapping", collapse=False)

        sls_mni = []
        self.logger.info("Combining Bundles...")
        for ii in tqdm(range(len(self.valid_ses_list))):
            this_sub = self.valid_sub_list[ii]
            this_ses = self.valid_ses_list[ii]
            seg_sft = aus.SegmentedSFT.fromfile(bundles_dict[
                this_sub][this_ses])
            seg_sft.sft.to_vox()
            sls = seg_sft.get_bundle(bundle_name).streamlines
            mapping = mapping_dict[this_sub][this_ses]

            if len(sls) > 0:
                delta = dts.values_from_volume(
                    mapping.forward,
                    sls, np.eye(4))
                sls_mni.extend([d + s for d, s in zip(delta, sls)])

        moved_sft = StatefulTractogram(
            sls_mni,
            reg_template,
            Space.VOX)
        save_tractogram(
            moved_sft,
            op.abspath(op.join(
                self.afq_path,
                f"bundle-{bundle_name}_subjects-all_MNI.trk")),
            bbox_valid_check=False)

    def upload_to_s3(self, s3fs, remote_path):
        """ Upload entire AFQ derivatives folder to S3"""
        s3fs.put(self.afq_path, remote_path, recursive=True)
        if op.exists(self.afqb_path):
            s3fs.put(self.afqb_path, remote_path, recursive=True)

    def export_group_density(self, boolify=True):
        """
        Generate a group density map by combining single subject density maps.

        Parameters
        ----------
        boolify : bool
            Whether to turn subject streamline count images into booleans
            before adding them into the group density map.

        Return
        ------
        Path to density nifti file.
        """
        densities = self.export("density_maps", collapse=False)
        ex_density_init = nib.load(densities[
            self.valid_sub_list[0]][
                self.valid_ses_list[0]])  # for shape and header

        group_density = np.zeros_like(ex_density_init.get_fdata())
        self.logger.info("Generating Group Density...")
        for ii in tqdm(range(len(self.valid_ses_list))):
            this_sub = self.valid_sub_list[ii]
            this_ses = self.valid_ses_list[ii]
            this_density = nib.load(densities[this_sub][this_ses]).get_fdata()
            if boolify:
                this_density = this_density.astype(bool)

            group_density = group_density + this_density
        group_density = group_density / len(self.valid_sub_list)
        group_density = nib.Nifti1Image(
            group_density,
            ex_density_init.affine,
            header=ex_density_init.header
        )

        out_fname = op.abspath(op.join(
            self.afq_path,
            f"desc-density_subjects-all_space-MNI_dwi.nii.gz"))
        nib.save(group_density, out_fname)
        return out_fname

    def assemble_AFQ_browser(self, output_path=None, metadata=None,
                             page_title="AFQ Browser", page_subtitle="",
                             page_title_link="", page_subtitle_link=""):
        """
        Assembles an instance of the AFQ-Browser from this AFQ instance.
        First, we generate the combined tract profile if it is not already
        generated. This includes running the full AFQ pipeline if it has not
        already run. The combined tract profile is one of the outputs of
        export_all.
        Second, we generate a streamlines.json file from the bundle
        recognized in the first subject's first session.
        Third, we call AFQ-Browser's assemble to assemble an AFQ-Browser
        instance in output_path.

        Parameters
        ----------
        output_path : str
            Path to location to create this instance of the browser in.
            Called "target" in AFQ Browser API. If None,
            bids_path/derivatives/afq_browser is used.
            Default: None
        metadata : str
            Path to subject metadata csv file. If None, an metadata file
            containing only subject ID is created. This file requires a
            "subjectID" column to work.
            Default: None
        page_title : str
            Page title. If None, prompt is sent to command line.
            Default: "AFQ Browser"
        page_subtitle : str
            Page subtitle. If None, prompt is sent to command line.
            Default: ""
        page_title_link : str
            Title hyperlink (including http(s)://).
            If None, prompt is sent to command line.
            Default: ""
        page_subtitle_link : str
            Subtitle hyperlink (including http(s)://).
            If None, prompt is sent to command line.
            Default: ""
        """
        if not using_afqb:
            self.logger.warning((
                "AFQ Browser is not installed, so AFQ Browser instance "
                "cannot be assembled. AFQ Browser can be installed with: "
                "`pip install pyAFQ[afqbrowser]` or "
                "`pip install AFQ-Browser>=0.3`"))
            return
        n_points_profile = self.export("n_points_profile", collapse=False)[
            self.valid_sub_list[0]][
                self.valid_ses_list[0]]
        if n_points_profile != 100:
            self.logger.warning((
                "AFQ Browser requires 100 points per tract profile, "
                "so AFQ Browser instance cannot be assembled."))
            return

        if output_path is None:
            output_path = self.afqb_path
        os.makedirs(self.afqb_path, exist_ok=True)

        # generate combined profiles csv
        self.combine_profiles()

        # generate streamlines.json file
        sls_json_fname = self.get_streamlines_json()

        afqb.assemble(
            op.abspath(op.join(self.afq_path, "tract_profiles.csv")),
            target=output_path,
            metadata=metadata,
            streamlines=sls_json_fname,
            title=page_title,
            subtitle=page_subtitle,
            link=page_title_link,
            sublink=page_subtitle_link)


class ParallelGroupAFQ():
    def __init__(self, *args, **kwargs):
        orig = GroupAFQ(*args, **kwargs)

        orig.parallel_params["submitter_params"] = \
            orig.parallel_params.get("submitter_params", {"plugin": "cf"})

        orig.parallel_params["cache_dir"] = \
            orig.parallel_params.get("cache_dir", None)

        self.parallel_params = orig.parallel_params
        self.pAFQ_kwargs = orig.pAFQ_inputs_list

        self.finishing_params = dict()
        self.finishing_params["args"] = args
        self.finishing_params["kwargs"] = kwargs
        self.finishing_params["output_dirs"] = [pAFQ.kwargs["output_dir"]
                                                for pAFQ in orig.pAFQ_list]

    def _submit_pydra(self, runnable):
        try:
            with pydra.Submitter(
                **self.parallel_params["submitter_params"],
            ) as sub:
                sub(runnable=runnable)
        # Addresses https://github.com/nipype/pydra/issues/630
        except AttributeError as e:
            if "'NoneType' object has no attribute 'replace'" not in str(e):
                raise

    def export(self, attr_name="help", collapse=True):
        f"""
        Export a specific output. To print a list of available outputs,
        call export without arguments.
        {valid_exports_string}

        Parameters
        ----------
        attr_name : str
            Name of the output to export. Default: "help"
        collapse : bool
            Whether to collapse session dimension if there is only 1 session.
            Default: True

        Returns
        -------
        output : dict
            The specific output as a dictionary. Keys are subjects.
            Values are dictionaries with keys of sessions
            if multiple sessions are used. Otherwise, values are
            the output.
            None if called without arguments.
        """

        @pydra.mark.task
        def export_sub(pAFQ_kwargs, attr_name):
            pAFQ = ParticipantAFQ(
                pAFQ_kwargs.dwi_data_file,
                pAFQ_kwargs.bval_file,
                pAFQ_kwargs.bvec_file,
                pAFQ_kwargs.results_dir,
                **pAFQ_kwargs.kwargs)
            pAFQ.export(attr_name)

        # Submit to pydra
        export_sub_task = export_sub(
            attr_name=attr_name,
            cache_dir=self.parallel_params["cache_dir"]
        ).split("pAFQ_kwargs", pAFQ_kwargs=self.pAFQ_kwargs)
        self._submit_pydra(export_sub_task)

    def export_all(self, viz=True, afqbrowser=True, xforms=True, indiv=True):
        """ Exports all the possible outputs

        Parameters
        ----------
        viz : bool
            Whether to output visualizations. This includes tract profile
            plots, a figure containing all bundles, and, if using the AFQ
            segmentation algorithm, individual bundle figures.
            Default: True
        afqbrowser : bool
            Whether to output an AFQ-Browser from this AFQ instance.
            Default: True
        xforms : bool
            Whether to output the reg_template image in subject space and,
            depending on if it is possible based on the mapping used, to
            output the b0 in template space.
            Default: True
        indiv : bool
            Whether to output individual bundles in their own files, in
            addition to the one file containing all bundles. If using
            the AFQ segmentation algorithm, individual ROIs are also
            output.
            Default: True
        """
        @pydra.mark.task
        def export_sub(
            pAFQ_kwargs,
            finishing_params,
            viz,
            afqbrowser,
            xforms,
            indiv
        ):
            pAFQ = ParticipantAFQ(
                pAFQ_kwargs.dwi_data_file,
                pAFQ_kwargs.bval_file,
                pAFQ_kwargs.bvec_file,
                pAFQ_kwargs.results_dir,
                **pAFQ_kwargs.kwargs)
            pAFQ.export_all(viz, xforms, indiv)

            for dir in finishing_params["output_dirs"]:
                if not glob.glob(op.join(dir, "*_desc-profiles_dwi.csv")):
                    return

            gAFQ = GroupAFQ(*finishing_params["args"],
                            **finishing_params["kwargs"])
            gAFQ.export_all(viz, afqbrowser, xforms, indiv)

        # Submit to pydra
        export_sub_task = export_sub(
            finishing_params=self.finishing_params,
            viz=viz,
            afqbrowser=afqbrowser,
            xforms=xforms,
            indiv=indiv,
            cache_dir=self.parallel_params["cache_dir"]
        ).split("pAFQ_kwargs", pAFQ_kwargs=self.pAFQ_kwargs)
        self._submit_pydra(export_sub_task)


def download_and_combine_afq_profiles(bucket,
                                      study_s3_prefix="", deriv_name=None,
                                      out_file=None,
                                      upload=False, session=None,
                                      **kwargs):
    """
    Download and combine tract profiles from different subjects / sessions
    on an s3 bucket into one CSV.
    Parameters
    ----------
    bucket : str
        The S3 bucket that contains the study data.
    study_s3_prefix : str
        The S3 prefix common to all of the study objects on S3.
    out_file : filename, optional
        Filename for the combined output CSV.
    deriv_name : str, optional
        If deriv_name is not None, it should be a string that specifies
        which derivatives folder to download and combine profiles from.
    upload : bool or str, optional
        If True, upload the combined CSV to Amazon S3 at
        bucket/study_s3_prefix/derivatives/afq. If a string,
        assume string is an Amazon S3 URI and upload there.
        Defaut: False
    session : str, optional
        Session to get CSVs from. If None, all sessions are used.
        Default: None
    kwargs : optional
        Optional arguments to pass to S3BIDSStudy.
    Returns
    -------
    Ouput CSV's pandas dataframe.
    """
    if "subjects" not in kwargs:
        kwargs["subjects"] = "all"
    if "anon" not in kwargs:
        kwargs["anon"] = False
    if deriv_name is None:
        deriv_name = True

    with nib.tmpdirs.InTemporaryDirectory() as t_dir:
        remote_study = S3BIDSStudy(
            "get_profiles",
            bucket,
            study_s3_prefix,
            **kwargs)
        remote_study.download(
            t_dir,
            include_modality_agnostic=False,
            include_derivs=deriv_name,
            include_derivs_dataset_description=True,
            suffix="profiles.csv")
        temp_study = BIDSLayout(t_dir, validate=False, derivatives=True)
        if session is None:
            profiles = temp_study.get(
                extension='csv',
                suffix='profiles',
                return_type='filename')
        else:
            profiles = temp_study.get(
                session=session,
                extension='csv',
                suffix='profiles',
                return_type='filename')

        df = combine_list_of_profiles(profiles)
        df.to_csv("tmp.csv", index=False)
        if upload is True:
            bids_prefix = "/".join([bucket, study_s3_prefix]).rstrip("/")
            fs = s3fs.S3FileSystem()
            fs.put(
                "tmp.csv",
                "/".join([
                    bids_prefix,
                    "derivatives",
                    "afq",
                    "combined_tract_profiles.csv"
                ]))
        elif isinstance(upload, str):
            fs = s3fs.S3FileSystem()
            fs.put("tmp.csv", upload.replace("s3://", ""))

    if out_file is not None:
        out_file = op.abspath(out_file)
        os.makedirs(op.dirname(out_file), exist_ok=True)
        df = clean_pandas_df(df)
        df.to_csv(out_file, index=False)

    return df


def combine_list_of_profiles(profile_fnames):
    """
    Combine tract profiles from different subjects / sessions
    into one CSV.

    Parameters
    ----------
    profile_fnames : list of str
        List of csv filenames.

    Returns
    -------
    Ouput CSV's pandas dataframe.
    """
    dfs = []
    for fname in profile_fnames:
        profiles = pd.read_csv(fname)
        profiles['subjectID'] = fname.split('sub-')[1].split('/')[0]
        if 'ses-' in fname:
            session_name = fname.split('ses-')[1].split('/')[0]
        else:
            session_name = 'unknown'
        profiles['sessionID'] = session_name
        dfs.append(profiles)

    return clean_pandas_df(pd.concat(dfs))
